"""Sup3r exogenous data handling"""

import logging
import numpy as np
from warnings import warn

from sup3r.utilities.topo import TopoExtractH5, TopoExtractNC
import sup3r.preprocessing.data_handling
import sup3r.utilities.topo
from sup3r.utilities.utilities import get_source_type

logger = logging.getLogger(__name__)


class ExogenousDataHandler:
    """Class to extract exogenous features for multistep forward passes. e.g.
    Multiple topography arrays at different resolutions for multiple spatial
    enhancement steps."""

    def __init__(self, file_paths, features, source_file, s_enhancements,
                 agg_factors, target=None, shape=None, raster_file=None,
                 max_delta=20, input_handler=None, topo_handler=None,
                 exo_steps=None):
        """
        Parameters
        ----------
        file_paths : str | list
            A single source h5 file or netcdf file to extract raster data from.
            The string can be a unix-style file path which will be passed
            through glob.glob. This is typically low-res WRF output or GCM
            netcdf data that is source low-resolution data intended to be
            sup3r resolved.
        features : list
            List of exogenous features to extract from source_h5
        source_file : str
            Filepath to source wtk, nsrdb, or netcdf file to get hi-res (2km or
            4km) data from which will be mapped to the enhanced grid of the
            file_paths input
        s_enhancements : list
            List of factors by which the Sup3rGan model will enhance the
            spatial dimensions of low resolution data from file_paths input
            where the total spatial enhancement is the product of these
            factors. For example, if file_paths has 100km data and there are 2
            spatial enhancement steps of 4x and 5x to a nominal resolution of
            5km, s_enhancements should be [1, 4, 5] and exo_steps should be
            [0, 1, 2] so that the input to the 4x model gets exogenous data
            at 100km (s_enhance=1, exo_step=0), the input to the 5x model gets
            exogenous data at 25km (s_enhance=4, exo_step=1), and there is a
            20x (1*4*5) exogeneous data layer available if the second model can
            receive a high-res input feature (e.g. WindGan). The length of this
            list should be equal to the number of agg_factors and the number of
            exo_steps
        agg_factors : list
            List of factors by which to aggregate the topo_source_h5 elevation
            data to the resolution of the file_paths input enhanced by
            s_enhance. The length of this list should be equal to the number of
            s_enhancements and the number of exo_steps
        target : tuple
            (lat, lon) lower left corner of raster. Either need target+shape or
            raster_file.
        shape : tuple
            (rows, cols) grid size. Either need target+shape or raster_file.
        raster_file : str | None
            File for raster_index array for the corresponding target and shape.
            If specified the raster_index will be loaded from the file if it
            exists or written to the file if it does not yet exist.  If None
            raster_index will be calculated directly. Either need target+shape
            or raster_file.
        input_handler : str
            data handler class to use for input data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.
        topo_handler : str
            topo extract class to use for source data. Provide a string name to
            match a class in topo.py. If None the correct handler will
            be guessed based on file type and time series properties.
        exo_steps : list
            List of model step indices for which exogenous data is required.
            e.g. If we have two model steps which take exo data and one which
            does not exo_steps = [0, 1]. The length of this list should be
            equal to the number of s_enhancements and the number of agg_factors
        """

        self.features = features
        self.s_enhancements = s_enhancements
        self.agg_factors = agg_factors
        self.data = []
        exo_steps = exo_steps or np.arange(len(self.s_enhancements))

        if self.s_enhancements[0] != 1:
            msg = ('s_enhancements typically starts with 1 so the first '
                   'exogenous data input matches the spatial resolution of '
                   'the source low-res input data, but received '
                   's_enhancements: {}'.format(self.s_enhancements))
            logger.warning(msg)
            warn(msg)

        msg = ('Need to provide the same number of enhancement factors and '
               f'agg factors. Received s_enhancements={s_enhancements} and '
               f'agg_factors={agg_factors}.')
        assert len(self.s_enhancements) == len(self.agg_factors), msg

        msg = ('Need to provide an integer enhancement factor for each model'
               'step. If the step is temporal enhancement then s_enhance=1')
        assert not any(s is None for s in self.s_enhancements), msg

        for i in range(len(self.s_enhancements)):
            s_enhance = np.product(self.s_enhancements[:i + 1])
            agg_factor = self.agg_factors[i]
            fdata = []
            if i in exo_steps:
                for f in features:
                    if f == 'topography':
                        topo_handler = self.get_topo_handler(source_file,
                                                             topo_handler)
                        data = topo_handler(file_paths, source_file, s_enhance,
                                            agg_factor, target=target,
                                            shape=shape,
                                            raster_file=raster_file,
                                            max_delta=max_delta,
                                            input_handler=input_handler)
                        data = data.hr_elev
                        fdata.append(data)
                    else:
                        msg = (f"Can only extract topography. Recived {f}.")
                        raise NotImplementedError(msg)
                self.data.append(np.stack(fdata, axis=-1))
            else:
                self.data.append(None)

    @staticmethod
    def get_topo_handler(source_file, topo_handler):
        """Get topo extraction class for source file

        Parameters
        ----------
        source_file : str
            Filepath to source wtk, nsrdb, or netcdf file to get hi-res (2km or
            4km) data from which will be mapped to the enhanced grid of the
            file_paths input
        topo_handler : str
            topo extract class to use for source data. Provide a string name to
            match a class in topo.py. If None the correct handler will
            be guessed based on file type and time series properties.

        Returns
        -------
        topo_handler : str
            topo extract class to use for source data.
        """
        if topo_handler is None:
            in_type = get_source_type(source_file)
            if in_type == 'nc':
                topo_handler = TopoExtractNC
            elif in_type == 'h5':
                topo_handler = TopoExtractH5
            else:
                msg = ('Did not recognize input type "{}" for file paths: {}'
                       .format(in_type, source_file))
                logger.error(msg)
                raise RuntimeError(msg)
        elif isinstance(topo_handler, str):
            topo_handler = getattr(sup3r.utilities.topo, topo_handler, None)
            if topo_handler is None:
                msg = ('Could not find requested topo handler class '
                       f'"{topo_handler}" in '
                       'sup3r.utilities.topo.')
                logger.error(msg)
                raise KeyError(msg)

        return topo_handler



# -*- coding: utf-8 -*-
"""sup3r batch utilities based on reV's batch module"""
from sup3r.pipeline.pipeline import Sup3rPipeline
from sup3r.pipeline.pipeline_cli import pipeline_monitor_background
from reV.batch.batch import BatchJob as RevBatchJob


class BatchJob(RevBatchJob):
    """Framework for building a batched job suite."""

    # Class attributes to set the software's pipeline class and run+monitor in
    # background method. These can be updated in dependent software packages
    # for other workflows that want to utilize reV's pipeline and batch
    # features.
    PIPELINE_CLASS = Sup3rPipeline
    PIPELINE_BACKGROUND_METHOD = pipeline_monitor_background



# -*- coding: utf-8 -*-
"""Bias calculation and correction modules."""
from .bias_transforms import (global_linear_bc, local_linear_bc,
                              monthly_local_linear_bc)
from .bias_calc import LinearCorrection, MonthlyLinearCorrection



# -*- coding: utf-8 -*-
"""Module to plot feature output from forward passes for visual inspection"""
import numpy as np
import matplotlib.pyplot as plt
import logging
import glob
from datetime import datetime as dt
import os

import rex
from rex.utilities.fun_utils import get_fun_call_str
from concurrent.futures import ThreadPoolExecutor, as_completed

from sup3r.utilities import ModuleName
from sup3r.utilities.cli import BaseCLI


logger = logging.getLogger(__name__)


class Sup3rVisualQa:
    """Module to plot features for visual qa"""

    def __init__(self, file_paths, out_pattern, features, time_step=10,
                 spatial_slice=slice(None), source_handler_class=None,
                 max_workers=None, overwrite=False, **kwargs):
        """
        Parameters
        ----------
        file_paths : list | str
            Specifies the files to use for the plotting routine. This is either
            a list of h5 files generated by the forward pass module or a string
            pointing to h5 forward pass output which can be parsed by glob.glob
        out_pattern : str
            The pattern to use for naming the plot figures. This must include
            {feature} and {index} so output files can be named with
            out_pattern.format(feature=feature, index=index).
            e.g. outfile_{feature}_{index}.png. The number of plot figures is
            determined by the time_index of the h5 files and the time_step
            argument. The index key refers to the plot file index from the list
            of all plot files generated.
        features : list
            List of features to plot from the h5 files provided.
        time_step : int
            Number of timesteps to average over for a single plot figure.
        spatial_slice : slice
            Slice specifying the spatial range to plot. This can include a
            step > 1 to speed up plotting.
        source_handler_class : str | None
            Name of the class to use for h5 input files. If None this defaults
            to MultiFileResource.
        max_workers : int | None
            Max number of workers to use for plotting. If workers=1 then all
            plots will be created in serial.
        overwrite : bool
            Whether to overwrite saved plots.
        **kwargs : dict
            Dictionary of kwargs passed to matplotlib.pyplot.scatter().
        """

        self.features = features
        self.out_pattern = out_pattern
        self.time_step = time_step
        self.spatial_slice = (spatial_slice if isinstance(spatial_slice, slice)
                              else slice(*spatial_slice))
        self.file_paths = (file_paths if isinstance(file_paths, list)
                           else glob.glob(file_paths))
        self.max_workers = max_workers
        self.kwargs = kwargs
        self.res_handler = source_handler_class or 'MultiFileResource'
        self.res_handler = getattr(rex, self.res_handler)
        self.overwrite = overwrite
        if not os.path.exists(os.path.dirname(out_pattern)):
            os.makedirs(os.path.dirname(out_pattern), exist_ok=True)
        logger.info('Initializing Sup3rVisualQa with '
                    f'file_paths={self.file_paths}, '
                    f'out_pattern={self.out_pattern}, '
                    f'features={self.features}, '
                    f'time_step={self.time_step}, '
                    f'spatial_slice={self.spatial_slice}, '
                    f'source_handler_class={self.res_handler}, '
                    f'max_workers={max_workers}, '
                    f'overwrite={self.overwrite}, '
                    f'kwargs={kwargs}.')

    def run(self):
        """
        Create plot figures for all the features in self.features. For each
        feature there will be n_files created, where n_files is the number of
        timesteps in the h5 files provided divided by self.time_step.
        """
        with self.res_handler(self.file_paths) as res:
            time_index = res.time_index
            n_files = len(time_index[::self.time_step])
            time_slices = np.array_split(np.arange(len(time_index)), n_files)
            time_slices = [slice(s[0], s[-1] + 1) for s in time_slices]

            if self.max_workers == 1:
                self._serial_figure_plots(res, time_index, time_slices,
                                          self.spatial_slice)
            else:
                self._parallel_figure_plots(res, time_index, time_slices,
                                            self.spatial_slice)

    def _serial_figure_plots(self, res, time_index, time_slices,
                             spatial_slice):
        """Plot figures in parallel with max_workers=self.workers

        Parameters
        ----------
        res : MultiFileResourceX
            Resource handler for the provided h5 files
        time_index : pd.DateTimeIndex
            The time index for the provided h5 files
        time_slices : list
             List of slices specifying all the time ranges to average and plot
        spatial_slice : slice
             Slice specifying the spatial range to plot
        """
        for feature in self.features:
            for i, t_slice in enumerate(time_slices):
                out_file = self.out_pattern.format(feature=feature,
                                                   index=str(i).zfill(8))
                self.plot_figure(res, time_index, feature, t_slice,
                                 spatial_slice, out_file)

    def _parallel_figure_plots(self, res, time_index, time_slices,
                               spatial_slice):
        """Plot figures in parallel with max_workers=self.workers

        Parameters
        ----------
        res : MultiFileResourceX
            Resource handler for the provided h5 files
        time_index : pd.DateTimeIndex
            The time index for the provided h5 files
        time_slices : list
             List of slices specifying all the time ranges to average and plot
        spatial_slice : slice
             Slice specifying the spatial range to plot
        """
        futures = {}
        now = dt.now()
        n_files = len(time_slices) * len(self.features)
        with ThreadPoolExecutor(max_workers=self.max_workers) as exe:
            for feature in self.features:
                for i, t_slice in enumerate(time_slices):
                    out_file = self.out_pattern.format(feature=feature,
                                                       index=str(i).zfill(8))
                    future = exe.submit(self.plot_figure, res, time_index,
                                        feature, t_slice, spatial_slice,
                                        out_file)
                    futures[future] = out_file

            logger.info(f'Started plotting {n_files} files '
                        f'in {dt.now() - now}.')

            for i, future in enumerate(as_completed(futures)):
                try:
                    future.result()
                except Exception as e:
                    msg = (f'Error making plot {futures[future]}.')
                    logger.exception(msg)
                    raise RuntimeError(msg) from e
                logger.debug(f'{i+1} out of {n_files} plots created.')

    def plot_figure(self, res, time_index, feature, t_slice, s_slice,
                    out_file):
        """Plot temporal average for the given feature and with the time range
        specified by t_slice

        Parameters
        ----------
        res : MultiFileResourceX
            Resource handler for the provided h5 files
        time_index : pd.DateTimeIndex
            The time index for the provided h5 files
        feature : str
            The feature to plot
        t_slice : slice
            The slice specifying the time range to average and plot
        s_slice : slice
            The slice specifying the spatial range to plot.
        out_file : str
            Name of the output plot file
        """
        if not self.overwrite and os.path.exists(out_file):
            logger.info(f'{out_file} already exists and overwrite='
                        f'{self.overwrite}. Skipping this plot.')
            return
        start_time = time_index[t_slice.start]
        stop_time = time_index[t_slice.stop - 1]
        logger.info(f'Plotting time average for {feature} from '
                    f'{start_time} to {stop_time}.')
        fig = plt.figure()
        title = f'{feature}: {start_time} - {stop_time}'
        plt.suptitle(title)
        plt.scatter(res.meta.longitude[s_slice], res.meta.latitude[s_slice],
                    c=np.mean(res[feature, t_slice, s_slice], axis=0),
                    **self.kwargs)
        plt.colorbar()
        fig.savefig(out_file)
        plt.close()
        logger.info(f'Saved figure {out_file}')

    @classmethod
    def get_node_cmd(cls, config):
        """Get a CLI call to initialize Sup3rVisualQa and execute the
        Sup3rVisualQa.run() method based on an input config

        Parameters
        ----------
        config : dict
            sup3r QA config with all necessary args and kwargs to
            initialize Sup3rVisualQa and execute Sup3rVisualQa.run()
        """
        import_str = 'import time;\n'
        import_str += 'from reV.pipeline.status import Status;\n'
        import_str += 'from rex import init_logger;\n'
        import_str += 'from sup3r.qa.visual_qa import Sup3rVisualQa;\n'

        qa_init_str = get_fun_call_str(cls, config)

        log_file = config.get('log_file', None)
        log_level = config.get('log_level', 'INFO')

        log_arg_str = (f'"sup3r", log_level="{log_level}"')
        if log_file is not None:
            log_arg_str += f', log_file="{log_file}"'

        cmd = (f"python -c \'{import_str}\n"
               "t0 = time.time();\n"
               f"logger = init_logger({log_arg_str});\n"
               f"qa = {qa_init_str};\n"
               "qa.run();\n"
               "t_elap = time.time() - t0;\n")

        cmd = BaseCLI.add_status_cmd(config, ModuleName.VISUAL_QA, cmd)
        cmd += (";\'\n")

        return cmd.replace('\\', '/')



# -*- coding: utf-8 -*-
"""
Abstract class to define the required interface for Sup3r model subclasses
"""
from abc import ABC, abstractmethod
import os
import time
import json
from concurrent.futures import ThreadPoolExecutor
from phygnn import CustomNetwork
from phygnn.layers.custom_layers import Sup3rAdder, Sup3rConcat
from rex.utilities.utilities import safe_json_load
import tensorflow as tf
from tensorflow.keras import optimizers
import numpy as np
import logging
import pprint
from warnings import warn

from sup3r.utilities import VERSION_RECORD
import sup3r.utilities.loss_metrics

logger = logging.getLogger(__name__)


class AbstractInterface(ABC):
    """
    Abstract class to define the required interface for Sup3r model subclasses

    Note that this only sets the required interfaces for a GAN that can be
    loaded from disk and used to predict synthetic outputs. The interface for
    models that can be trained will be set in another class.
    """

    @classmethod
    @abstractmethod
    def load(cls, model_dir, verbose=True):
        """Load the GAN with its sub-networks from a previously saved-to output
        directory.

        Parameters
        ----------
        model_dir
            Directory to load GAN model files from.
        verbose : bool
            Flag to log information about the loaded model.

        Returns
        -------
        out : BaseModel
            Returns a pretrained gan model that was previously saved to
            model_dir
        """

    @staticmethod
    def seed(s=0):
        """
        Set the random seed for reproducible results.

        Parameters
        ----------
        s : int
            Random seed
        """
        CustomNetwork.seed(s=s)

    @abstractmethod
    def generate(self, low_res):
        """Use the generator model to generate high res data from low res
        input. This is the public generate function.

        Parameters
        ----------
        low_res : np.ndarray
            Low-resolution input data, usually a 4D or 5D array of shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)

        Returns
        -------
        hi_res : ndarray
            Synthetically generated high-resolution data, usually a 4D or 5D
            array with shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        """

    @property
    def s_enhance(self):
        """Factor by which model will enhance spatial resolution. Used in
        model training during high res coarsening"""
        return self.meta.get('s_enhance', None)

    @property
    def t_enhance(self):
        """Factor by which model will enhance temporal resolution. Used in
        model training during high res coarsening"""
        return self.meta.get('t_enhance', None)

    @property
    @abstractmethod
    def meta(self):
        """Get meta data dictionary that defines how the model was created"""

    @property
    def training_features(self):
        """Get the list of input feature names that the generative model was
        trained on."""
        return self.meta.get('training_features', None)

    @property
    def output_features(self):
        """Get the list of output feature names that the generative model
        outputs and that the discriminator predicts on."""
        return self.meta.get('output_features', None)

    @property
    def smoothing(self):
        """Value of smoothing parameter used in gaussian filtering of coarsened
        high res data."""
        return self.meta.get('smoothing', None)

    @property
    def smoothed_features(self):
        """Get the list of smoothed input feature names that the generative
        model was trained on."""
        return self.meta.get('smoothed_features', None)

    @property
    def model_params(self):
        """
        Model parameters, used to save model to disc

        Returns
        -------
        dict
        """
        model_params = {'meta': self.meta}
        return model_params

    @property
    def version_record(self):
        """A record of important versions that this model was built with.

        Returns
        -------
        dict
        """
        return VERSION_RECORD

    def set_model_params(self, **kwargs):
        """Set parameters used for training the model

        Parameters
        ----------
        kwargs : dict
            Keyword arguments including 'training_features', 'output_features',
            'smoothed_features', 's_enhance', 't_enhance', 'smoothing'
        """

        keys = ('training_features', 'output_features', 'smoothed_features',
                's_enhance', 't_enhance', 'smoothing')
        keys = [k for k in keys if k in kwargs]

        for var in keys:
            val = getattr(self, var, None)
            if val is None:
                self.meta[var] = kwargs[var]
            elif val != kwargs[var]:
                msg = ('Model was previously trained with {var}={} but '
                       'received new {var}={}'
                       .format(val, kwargs[var], var=var))
                logger.warning(msg)
                warn(msg)

    def save_params(self, out_dir):
        """
        Parameters
        ----------
        out_dir : str
            Directory to save linear model params. This directory will be
            created if it does not already exist.
        """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        fp_params = os.path.join(out_dir, 'model_params.json')
        with open(fp_params, 'w') as f:
            params = self.model_params
            json.dump(params, f, sort_keys=True, indent=2)


# pylint: disable=E1101,W0201,E0203
class AbstractSingleModel(ABC):
    """
    Abstract class to define the required training interface
    for Sup3r model subclasses
    """

    def __init__(self):
        self.gpu_list = tf.config.list_physical_devices('GPU')
        self.default_device = '/cpu:0'
        self._version_record = VERSION_RECORD
        self.name = None
        self._meta = None
        self.loss_name = None
        self.loss_fun = None
        self._history = None
        self._optimizer = None
        self._gen = None
        self._means = None
        self._stdevs = None

    def load_network(self, model, name):
        """Load a CustomNetwork object from hidden layers config, .json file
        config, or .pkl file saved pre-trained model.

        Parameters
        ----------
        model : str | dict
            Model hidden layers config, a .json with "hidden_layers" key, or a
            .pkl for a saved pre-trained model.
        name : str
            Name of the model to be loaded

        Returns
        -------
        model : phygnn.CustomNetwork
            CustomNetwork object initialized from the model input.
        """

        if isinstance(model, str) and model.endswith('.json'):
            model = safe_json_load(model)
            self._meta[f'config_{name}'] = model
            if 'hidden_layers' in model:
                model = model['hidden_layers']
            elif ('meta' in model
                  and f'config_{name}' in model['meta']
                  and 'hidden_layers' in model['meta'][f'config_{name}']):
                model = model['meta'][f'config_{name}']['hidden_layers']
            else:
                msg = ('Could not load model from json config, need '
                       '"hidden_layers" key or '
                       f'"meta/config_{name}/hidden_layers" '
                       ' at top level but only found: {}'
                       .format(model.keys()))
                logger.error(msg)
                raise KeyError(msg)

        elif isinstance(model, str) and model.endswith('.pkl'):
            with tf.device(self.default_device):
                model = CustomNetwork.load(model)

        if isinstance(model, list):
            model = CustomNetwork(hidden_layers=model, name=name)

        if not isinstance(model, CustomNetwork):
            msg = ('Something went wrong. Tried to load a custom network '
                   'but ended up with a model of type "{}"'
                   .format(type(model)))
            logger.error(msg)
            raise TypeError(msg)

        return model

    @property
    def means(self):
        """Get the data normalization mean values.

        Returns
        -------
        np.ndarray
        """
        return self._means

    @property
    def stdevs(self):
        """Get the data normalization standard deviation values.

        Returns
        -------
        np.ndarray
        """
        return self._stdevs

    @property
    def output_stdevs(self):
        """Get the data normalization standard deviation values for only the
        output features
        Returns
        -------
        np.ndarray
        """
        indices = [self.training_features.index(f)
                   for f in self.output_features]
        return self._stdevs[indices]

    @property
    def output_means(self):
        """Get the data normalization mean values for only the output features
        Returns
        -------
        np.ndarray
        """
        indices = [self.training_features.index(f)
                   for f in self.output_features]
        return self._means[indices]

    def set_norm_stats(self, new_means, new_stdevs):
        """Set the normalization statistics associated with a data batch
        handler to model attributes.

        Parameters
        ----------
        new_means : list | tuple | np.ndarray
            1D iterable of mean values with same length as number of features.
        new_stdevs : list | tuple | np.ndarray
            1D iterable of stdev values with same length as number of features.
        """

        if self._means is not None:
            logger.info('Setting new normalization statistics...')
            logger.info("Model's previous data mean values: {}"
                        .format(self._means))
            logger.info("Model's previous data stdev values: {}"
                        .format(self._stdevs))

        self._means = new_means
        self._stdevs = new_stdevs

        if not isinstance(self._means, np.ndarray):
            self._means = np.array(self._means)
        if not isinstance(self._stdevs, np.ndarray):
            self._stdevs = np.array(self._stdevs)

        logger.info('Set data normalization mean values: {}'
                    .format(self._means))
        logger.info('Set data normalization stdev values: {}'
                    .format(self._stdevs))

    def norm_input(self, low_res):
        """Normalize low resolution data being input to the generator.

        Parameters
        ----------
        low_res : np.ndarray
            Un-normalized low-resolution input data in physical units, usually
            a 4D or 5D array of shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)

        Returns
        -------
        low_res : np.ndarray
            Normalized low-resolution input data, usually a 4D or 5D array of
            shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        """
        if self._means is not None:
            if isinstance(low_res, tf.Tensor):
                low_res = low_res.numpy()

            if any(self._stdevs == 0):
                stdevs = np.where(self._stdevs == 0, 1, self._stdevs)
                msg = ('Some standard deviations are zero.')
                logger.warning(msg)
                warn(msg)
            else:
                stdevs = self._stdevs

            low_res = (low_res.copy() - self._means) / stdevs

        return low_res

    def un_norm_output(self, output):
        """Un-normalize synthetically generated output data to physical units

        Parameters
        ----------
        output : tf.Tensor | np.ndarray
            Synthetically generated high-resolution data

        Returns
        -------
        output : np.ndarray
            Synthetically generated high-resolution data
        """
        if self._means is not None:
            if isinstance(output, tf.Tensor):
                output = output.numpy()

            output = (output * self.output_stdevs) + self.output_means

        return output

    @property
    def optimizer(self):
        """Get the tensorflow optimizer to perform gradient descent
        calculations for the generative network. This is functionally identical
        to optimizer_disc is no special optimizer model or learning rate was
        specified for the disc.

        Returns
        -------
        tf.keras.optimizers.Optimizer
        """
        return self._optimizer

    @property
    def history(self):
        """
        Model training history DataFrame (None if not yet trained)

        Returns
        -------
        pandas.DataFrame | None
        """
        return self._history

    @property
    def generator(self):
        """Get the generative model.

        Returns
        -------
        phygnn.base.CustomNetwork
        """
        return self._gen

    @property
    def generator_weights(self):
        """Get a list of layer weights and bias terms for the generator model.

        Returns
        -------
        list
        """
        return self.generator.weights

    def _needs_lr_exo(self, low_res):
        """Determine whether or not the sup3r model needs low-res exogenous
        data

        Parameters
        ----------
        low_res : np.ndarray
            Low-resolution input data, usually a 4D or 5D array of shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)

        Returns
        -------
        needs_lr_exo : bool
            True if the model requires low-resolution exogenous data.
        """

        return low_res.shape[-1] < len(self.training_features)

    @staticmethod
    def init_optimizer(optimizer, learning_rate):
        """Initialize keras optimizer object.

        Parameters
        ----------
        optimizer : tf.keras.optimizers.Optimizer | dict | None | str
            Instantiated tf.keras.optimizers object or a dict optimizer config
            from tf.keras.optimizers.get_config(). None defaults to Adam.
        learning_rate : float, optional
            Optimizer learning rate. Not used if optimizer input arg is a
            pre-initialized object or if optimizer input arg is a config dict.

        Returns
        -------
        optimizer : tf.keras.optimizers.Optimizer
            Initialized optimizer object.
        """
        if isinstance(optimizer, dict):
            class_name = optimizer['name']
            OptimizerClass = getattr(optimizers, class_name)
            optimizer = OptimizerClass.from_config(optimizer)
        elif optimizer is None:
            optimizer = optimizers.Adam(learning_rate=learning_rate)

        return optimizer

    @staticmethod
    def load_saved_params(out_dir, verbose=True):
        """Load saved model_params (you need this and the gen+disc models
        to load a full model).

        Parameters
        ----------
        out_dir : str
            Directory to load model files from.
        verbose : bool
            Flag to log information about the loaded model.

        Returns
        -------
        params : dict
            Model parameters loaded from disk json file. This should be the
            same as self.model_params with and additional 'history' entry.
            Should be all the kwargs you need to init a model.
        """

        fp_params = os.path.join(out_dir, 'model_params.json')
        with open(fp_params, 'r') as f:
            params = json.load(f)

        # using the saved model dir makes this more portable
        fp_history = os.path.join(out_dir, 'history.csv')
        if os.path.exists(fp_history):
            params['history'] = fp_history
        else:
            params['history'] = None

        if 'version_record' in params:
            version_record = params.pop('version_record')
            if verbose:
                logger.info('Loading model from disk '
                            'that was created with the '
                            'following package versions: \n{}'
                            .format(pprint.pformat(version_record, indent=2)))

        return params

    @staticmethod
    def get_loss_fun(loss):
        """Get the initialized loss function class from the sup3r loss library
        or the tensorflow losses.

        Parameters
        ----------
        loss : str
            Loss function class name from sup3r.utilities.loss_metrics
            (prioritized) or tensorflow.keras.losses. Defaults to
            tf.keras.losses.MeanSquaredError.

        Returns
        -------
        out : tf.keras.losses.Loss
            Initialized loss function class that is callable, e.g. if
            "MeanSquaredError" is requested, this will return
            an instance of tf.keras.losses.MeanSquaredError()
        """

        out = getattr(sup3r.utilities.loss_metrics, loss, None)
        if out is None:
            out = getattr(tf.keras.losses, loss, None)

        if out is None:
            msg = ('Could not find requested loss function "{}" in '
                   'sup3r.utilities.loss_metrics or tf.keras.losses.'
                   .format(loss))
            logger.error(msg)
            raise KeyError(msg)

        return out()

    @staticmethod
    def get_optimizer_config(optimizer):
        """Get a config that defines the current model optimizer

        Parameters
        ----------
        optimizer : tf.keras.optimizers.Optimizer
            TF-Keras optimizer object

        Returns
        -------
        config : dict
            Optimizer config
        """
        conf = optimizer.get_config()
        for k, v in conf.items():
            # need to convert numpy dtypes to float/int for json.dump()
            if np.issubdtype(type(v), np.floating):
                conf[k] = float(v)
            elif np.issubdtype(type(v), np.integer):
                conf[k] = int(v)
        return conf

    @staticmethod
    def update_loss_details(loss_details, new_data, batch_len, prefix=None):
        """Update a dictionary of loss_details with loss information from a new
        batch.

        Parameters
        ----------
        loss_details : dict
            Namespace of the breakdown of loss components where each value is a
            running average at the current state in the epoch.
        new_data : dict
            Namespace of the breakdown of loss components for a single new
            batch.
        batch_len : int
            Length of the incomming batch.
        prefix : None | str
            Option to prefix the names of the loss data when saving to the
            loss_details dictionary.

        Returns
        -------
        loss_details : dict
            Same as input loss_details but with running averages updated.
        """
        assert 'n_obs' in loss_details, 'loss_details must have n_obs to start'
        prior_n_obs = loss_details['n_obs']
        new_n_obs = prior_n_obs + batch_len

        for key, new_value in new_data.items():
            key = key if prefix is None else prefix + key
            new_value = (new_value if not isinstance(new_value, tf.Tensor)
                         else new_value.numpy())

            if key in loss_details:
                saved_value = loss_details[key]
                saved_value *= prior_n_obs
                saved_value += batch_len * new_value
                saved_value /= new_n_obs
                loss_details[key] = saved_value
            else:
                loss_details[key] = new_value

        loss_details['n_obs'] = new_n_obs

        return loss_details

    @staticmethod
    def log_loss_details(loss_details, level='INFO'):
        """Log the loss details to the module logger.

        Parameters
        ----------
        loss_details : dict
            Namespace of the breakdown of loss components where each value is a
            running average at the current state in the epoch.
        level : str
            Log level (e.g. INFO, DEBUG)
        """
        for k, v in sorted(loss_details.items()):
            if k != 'n_obs':
                if isinstance(v, str):
                    msg_format = '\t{}: {}'
                else:
                    msg_format = '\t{}: {:.2e}'
                if level.lower() == 'info':
                    logger.info(msg_format.format(k, v))
                else:
                    logger.debug(msg_format.format(k, v))

    @staticmethod
    def early_stop(history, column, threshold=0.005, n_epoch=5):
        """Determine whether to stop training early based on nearly no change
        to validation loss for a certain number of consecutive epochs.

        Parameters
        ----------
        history : pd.DataFrame | None
            Model training history
        column : str
            Column from the model training history to evaluate for early
            termination.
        threshold : float
            The absolute relative fractional difference in validation loss
            between subsequent epochs below which an early termination is
            warranted. E.g. if val losses were 0.1 and 0.0998 the relative
            diff would be calculated as 0.0002 / 0.1 = 0.002 which would be
            less than the default thresold of 0.01 and would satisfy the
            condition for early termination.
        n_epoch : int
            The number of consecutive epochs that satisfy the threshold that
            warrants an early stop.

        Returns
        -------
        stop : bool
            Flag to stop training (True) or keep going (False).
        """
        stop = False

        if history is not None and len(history) > n_epoch + 1:
            diffs = np.abs(np.diff(history[column]))
            if all(diffs[-n_epoch:] < threshold):
                stop = True
                logger.info('Found early stop condition, loss values "{}" '
                            'have absolute relative differences less than '
                            'threshold {}: {}'
                            .format(column, threshold, diffs[-n_epoch:]))

        return stop

    @abstractmethod
    def save(self, low_res):
        """Save the model with its sub-networks to a directory.

        Parameters
        ----------
        out_dir : str
            Directory to save model files. This directory will be created
            if it does not already exist.
        """

    def finish_epoch(self, epoch, epochs, t0, loss_details,
                     checkpoint_int, out_dir,
                     early_stop_on, early_stop_threshold,
                     early_stop_n_epoch, extras=None):
        """Perform finishing checks after an epoch is done training

        Parameters
        ----------
        epoch : int
            Epoch number that is finishing
        epochs : list
            List of epochs being iterated through
        t0 : float
            Starting time of training.
        loss_details : dict
            Namespace of the breakdown of loss components
        checkpoint_int : int | None
            Epoch interval at which to save checkpoint models.
        out_dir : str
            Directory to save checkpoint models. Should have {epoch} in
            the directory name. This directory will be created if it does not
            already exist.
        early_stop_on : str | None
            If not None, this should be a column in the training history to
            evaluate for early stopping (e.g. validation_loss_gen,
            validation_loss_disc). If this value in this history decreases by
            an absolute fractional relative difference of less than 0.01 for
            more than 5 epochs in a row, the training will stop early.
        early_stop_threshold : float
            The absolute relative fractional difference in validation loss
            between subsequent epochs below which an early termination is
            warranted. E.g. if val losses were 0.1 and 0.0998 the relative
            diff would be calculated as 0.0002 / 0.1 = 0.002 which would be
            less than the default thresold of 0.01 and would satisfy the
            condition for early termination.
        early_stop_n_epoch : int
            The number of consecutive epochs that satisfy the threshold that
            warrants an early stop.
        extras : dict | None
            Extra kwargs/parameters to save in the epoch history.

        Returns
        -------
        stop : bool
            Flag to early stop training.
        """

        self.log_loss_details(loss_details)

        self._history.at[epoch, 'elapsed_time'] = time.time() - t0
        for key, value in loss_details.items():
            if key != 'n_obs':
                self._history.at[epoch, key] = value

        last_epoch = epoch == epochs[-1]
        chp = checkpoint_int is not None and (epoch % checkpoint_int) == 0
        if last_epoch or chp:
            msg = ('Model output dir for checkpoint models should have '
                   f'{"{epoch}"} but did not: {out_dir}')
            assert '{epoch}' in out_dir, msg
            self.save(out_dir.format(epoch=epoch))

        stop = False
        if early_stop_on is not None and early_stop_on in self._history:
            stop = self.early_stop(self._history, early_stop_on,
                                   threshold=early_stop_threshold,
                                   n_epoch=early_stop_n_epoch)
            if stop:
                self.save(out_dir.format(epoch=epoch))

        if extras is not None:
            for k, v in extras.items():
                self._history.at[epoch, k] = v

        return stop

    @tf.function()
    def get_single_grad(self, low_res, hi_res_true, training_weights,
                        device_name=None, **calc_loss_kwargs):
        """Run gradient descent for one mini-batch of (low_res, hi_res_true),
        do not update weights, just return gradient details.

        Parameters
        ----------
        low_res : np.ndarray
            Real low-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        hi_res_true : np.ndarray
            Real high-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        training_weights : list
            A list of layer weights that are to-be-trained based on the
            current loss weight values.
        device_name : None | str
            Optional tensorflow device name for GPU placement. Note that if a
            GPU is available, variables will be placed on that GPU even if
            device_name=None.
        calc_loss_kwargs : dict
            Kwargs to pass to the self.calc_loss() method

        Returns
        -------
        grad : list
            a list or nested structure of Tensors (or IndexedSlices, or None,
            or CompositeTensor) representing the gradients for the
            training_weights
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        with tf.device(device_name):
            with tf.GradientTape(watch_accessed_variables=False) as tape:
                tape.watch(training_weights)

                hi_res_gen = self._tf_generate(low_res)
                loss_out = self.calc_loss(hi_res_true, hi_res_gen,
                                          **calc_loss_kwargs)
                loss, loss_details = loss_out

                grad = tape.gradient(loss, training_weights)

        return grad, loss_details

    def run_gradient_descent(self, low_res, hi_res_true, training_weights,
                             optimizer=None, multi_gpu=False,
                             **calc_loss_kwargs):
        """Run gradient descent for one mini-batch of (low_res, hi_res_true)
        and update weights

        Parameters
        ----------
        low_res : np.ndarray
            Real low-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        hi_res_true : np.ndarray
            Real high-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        training_weights : list
            A list of layer weights that are to-be-trained based on the
            current loss weight values.
        optimizer : tf.keras.optimizers.Optimizer
            Optimizer class to use to update weights. This can be different if
            you're training just the generator or one of the discriminator
            models. Defaults to the generator optimizer.
        multi_gpu : bool
            Flag to break up the batch for parallel gradient descent
            calculations on multiple gpus. If True and multiple GPUs are
            present, each batch from the batch_handler will be divided up
            between the GPUs and the resulting gradient from each GPU will
            constitute a single gradient descent step with the nominal learning
            rate that the model was initialized with.
        calc_loss_kwargs : dict
            Kwargs to pass to the self.calc_loss() method

        Returns
        -------
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        t0 = time.time()
        if optimizer is None:
            optimizer = self.optimizer

        if not multi_gpu or len(self.gpu_list) == 1:
            grad, loss_details = self.get_single_grad(low_res, hi_res_true,
                                                      training_weights,
                                                      **calc_loss_kwargs)
            optimizer.apply_gradients(zip(grad, training_weights))
            t1 = time.time()
            logger.debug(f'Finished single gradient descent steps on '
                         f'{len(self.gpu_list)} GPUs in {(t1 - t0):.3f}s')

        else:
            futures = []
            lr_chunks = np.array_split(low_res, len(self.gpu_list))
            hr_true_chunks = np.array_split(hi_res_true, len(self.gpu_list))

            with ThreadPoolExecutor(max_workers=len(self.gpu_list)) as exe:
                for i in range(len(self.gpu_list)):
                    futures.append(exe.submit(self.get_single_grad,
                                              lr_chunks[i],
                                              hr_true_chunks[i],
                                              training_weights,
                                              device_name=f'/gpu:{i}',
                                              **calc_loss_kwargs))
            for i, future in enumerate(futures):
                grad, loss_details = future.result()
                optimizer.apply_gradients(zip(grad, training_weights))

            t1 = time.time()
            logger.debug(f'Finished {len(futures)} gradient descent steps on '
                         f'{len(self.gpu_list)} GPUs in {(t1 - t0):.3f}s')

        return loss_details


# pylint: disable=E1101,W0201,E0203
class AbstractWindInterface(ABC):
    """
    Abstract class to define the required training interface
    for Sup3r wind model subclasses
    """
    # pylint: disable=E0211
    @staticmethod
    def set_model_params(**kwargs):
        """Set parameters used for training the model

        Parameters
        ----------
        kwargs : dict
            Keyword arguments including 'training_features', 'output_features',
            'smoothed_features', 's_enhance', 't_enhance', 'smoothing'. For the
            Wind classes, the last entry in "output_features" must be
            "topography"

        Returns
        -------
        kwargs : dict
            Same as input but with topography removed from "output_features",
            this is because topography is concatenated mid-network in the
            WindGan generators and is not an output feature but is required in
            the hi-res training set.
        """
        output_features = kwargs['output_features']
        msg = ('Last output feature from the data handler must be topography '
               'to train the WindCC model, but received output features: {}'
               .format(output_features))
        assert output_features[-1] == 'topography', msg
        output_features.remove('topography')
        kwargs['output_features'] = output_features
        return kwargs

    def _reshape_norm_topo(self, hi_res, hi_res_topo, norm_in=True):
        """Reshape the hi_res_topo to match the hi_res tensor (if necessary)
        and normalize (if requested).

        Parameters
        ----------
        hi_res : ndarray
            Synthetically generated high-resolution data, usually a 4D or 5D
            array with shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        hi_res_topo : np.ndarray
            This should be a 4D array for spatial enhancement model or 5D array
            for a spatiotemporal enhancement model (obs, spatial_1, spatial_2,
            (temporal), features) corresponding to the high-resolution
            spatial_1 and spatial_2. This data will be input to the custom
            phygnn Sup3rAdder or Sup3rConcat layer if found in the generative
            network. This differs from the exogenous_data input in that
            exogenous_data always matches the low-res input. For this function,
            hi_res_topo can also be a 2D array (spatial_1, spatial_2). Note
            that this input gets normalized if norm_in=True.
        norm_in : bool
            Flag to normalize low_res input data if the self._means,
            self._stdevs attributes are available. The generator should always
            received normalized data with mean=0 stdev=1. This also normalizes
            hi_res_topo.

        Returns
        -------
        hi_res_topo : np.ndarray
            Same as input but reshaped to match hi_res (if necessary) and
            normalized (if requested)
        """
        if hi_res_topo is None:
            return hi_res_topo

        if norm_in and self._means is not None:
            idf = self.training_features.index('topography')
            hi_res_topo = ((hi_res_topo.copy() - self._means[idf])
                           / self._stdevs[idf])

        if len(hi_res_topo.shape) > 2:
            slicer = [0] * len(hi_res_topo.shape)
            slicer[1] = slice(None)
            slicer[2] = slice(None)
            hi_res_topo = hi_res_topo[tuple(slicer)]

        if len(hi_res.shape) == 4:
            hi_res_topo = np.expand_dims(hi_res_topo, axis=(0, 3))
            hi_res_topo = np.repeat(hi_res_topo, hi_res.shape[0], axis=0)
        elif len(hi_res.shape) == 5:
            hi_res_topo = np.expand_dims(hi_res_topo, axis=(0, 3, 4))
            hi_res_topo = np.repeat(hi_res_topo, hi_res.shape[0], axis=0)
            hi_res_topo = np.repeat(hi_res_topo, hi_res.shape[3], axis=3)

        if len(hi_res_topo.shape) != len(hi_res.shape):
            msg = ('hi_res and hi_res_topo arrays are not of the same rank: '
                   '{} and {}'.format(hi_res.shape, hi_res_topo.shape))
            logger.error(msg)
            raise RuntimeError(msg)

        return hi_res_topo

    def generate(self, low_res, norm_in=True, un_norm_out=True,
                 exogenous_data=None):
        """Use the generator model to generate high res data from low res
        input. This is the public generate function.

        Parameters
        ----------
        low_res : np.ndarray
            Low-resolution input data, usually a 4D or 5D array of shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        norm_in : bool
            Flag to normalize low_res input data if the self._means,
            self._stdevs attributes are available. The generator should always
            received normalized data with mean=0 stdev=1. This also normalizes
            hi_res_topo.
        un_norm_out : bool
           Flag to un-normalize synthetically generated output data to physical
           units
        exogenous_data : ndarray | list | None
            Exogenous data for topography inputs. The first entry in this list
            (or only entry) is a low-resolution topography array that can be
            concatenated to the low_res input array. The second entry is
            high-resolution topography (either 2D or 4D/5D depending on if
            spatial or spatiotemporal super res).

        Returns
        -------
        hi_res : ndarray
            Synthetically generated high-resolution data, usually a 4D or 5D
            array with shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        """
        low_res_topo = None
        hi_res_topo = None
        if isinstance(exogenous_data, np.ndarray):
            low_res_topo = exogenous_data
        elif isinstance(exogenous_data, (list, tuple)):
            low_res_topo = exogenous_data[0]
            if len(exogenous_data) > 1:
                hi_res_topo = exogenous_data[1]

        exo_check = (low_res is None or not self._needs_lr_exo(low_res))
        low_res = (low_res if exo_check
                   else np.concatenate((low_res, low_res_topo), axis=-1))

        if norm_in and self._means is not None:
            low_res = self.norm_input(low_res)

        hi_res = self.generator.layers[0](low_res)
        for i, layer in enumerate(self.generator.layers[1:]):
            try:
                if (isinstance(layer, (Sup3rAdder, Sup3rConcat))
                        and hi_res_topo is not None):
                    hi_res_topo = self._reshape_norm_topo(hi_res, hi_res_topo,
                                                          norm_in=norm_in)
                    hi_res = layer(hi_res, hi_res_topo)
                else:
                    hi_res = layer(hi_res)
            except Exception as e:
                msg = ('Could not run layer #{} "{}" on tensor of shape {}'
                       .format(i + 1, layer, hi_res.shape))
                logger.error(msg)
                raise RuntimeError(msg) from e

        hi_res = hi_res.numpy()

        if un_norm_out and self._means is not None:
            hi_res = self.un_norm_output(hi_res)

        return hi_res

    @tf.function
    def _tf_generate(self, low_res, hi_res_topo):
        """Use the generator model to generate high res data from los res input

        Parameters
        ----------
        low_res : np.ndarray
            Real low-resolution data. The generator should always
            received normalized data with mean=0 stdev=1.
        hi_res_topo : np.ndarray
            This should be a 4D array for spatial enhancement model or 5D array
            for a spatiotemporal enhancement model (obs, spatial_1, spatial_2,
            (temporal), features) corresponding to the high-resolution
            spatial_1 and spatial_2. This data will be input to the custom
            phygnn Sup3rAdder or Sup3rConcat layer if found in the generative
            network. This differs from the exogenous_data input in that
            exogenous_data always matches the low-res input.

        Returns
        -------
        hi_res : tf.Tensor
            Synthetically generated high-resolution data
        """
        hi_res = self.generator.layers[0](low_res)
        for i, layer in enumerate(self.generator.layers[1:]):
            try:
                if (isinstance(layer, (Sup3rAdder, Sup3rConcat))
                        and hi_res_topo is not None):
                    hi_res = layer(hi_res, hi_res_topo)
                else:
                    hi_res = layer(hi_res)
            except Exception as e:
                msg = ('Could not run layer #{} "{}" on tensor of shape {}'
                       .format(i + 1, layer, hi_res.shape))
                logger.error(msg)
                raise RuntimeError(msg) from e

        return hi_res

    @tf.function()
    def get_single_grad(self, low_res, hi_res_true, training_weights,
                        device_name=None, **calc_loss_kwargs):
        """Run gradient descent for one mini-batch of (low_res, hi_res_true),
        do not update weights, just return gradient details.

        Parameters
        ----------
        low_res : np.ndarray
            Real low-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        hi_res_true : np.ndarray
            Real high-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        training_weights : list
            A list of layer weights that are to-be-trained based on the
            current loss weight values.
        device_name : None | str
            Optional tensorflow device name for GPU placement. Note that if a
            GPU is available, variables will be placed on that GPU even if
            device_name=None.
        calc_loss_kwargs : dict
            Kwargs to pass to the self.calc_loss() method

        Returns
        -------
        grad : list
            a list or nested structure of Tensors (or IndexedSlices, or None,
            or CompositeTensor) representing the gradients for the
            training_weights
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        hi_res_topo = hi_res_true[..., -1:]

        with tf.device(device_name):
            with tf.GradientTape(watch_accessed_variables=False) as tape:
                tape.watch(training_weights)

                hi_res_gen = self._tf_generate(low_res, hi_res_topo)
                loss_out = self.calc_loss(hi_res_true, hi_res_gen,
                                          **calc_loss_kwargs)
                loss, loss_details = loss_out

                grad = tape.gradient(loss, training_weights)

        return grad, loss_details



# -*- coding: utf-8 -*-
"""
sup3r bias correction calculation CLI entry points.
"""
import copy
import click
import logging
import os

import sup3r.bias.bias_calc
from sup3r.utilities import ModuleName
from sup3r.version import __version__
from sup3r.utilities.cli import BaseCLI


logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, verbose):
    """Sup3r bias calc Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='sup3r bias correction calculation config .json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run sup3r bias correction calculation from a config file."""
    config = BaseCLI.from_config_preflight(ModuleName.BIAS_CALC, ctx,
                                           config_file, verbose)

    exec_kwargs = config.get('execution_control', {})
    hardware_option = exec_kwargs.pop('option', 'local')
    calc_class_name = config['bias_calc_class']
    BiasCalcClass = getattr(sup3r.bias.bias_calc, calc_class_name)
    basename = config['job_name']
    log_pattern = config.get('log_pattern', None)

    jobs = config['jobs']
    for i_node, job in enumerate(jobs):
        node_config = copy.deepcopy(job)
        node_config['status_dir'] = config['status_dir']
        node_config['log_file'] = (
            log_pattern if log_pattern is None
            else os.path.normpath(log_pattern.format(node_index=i_node)))
        name = ('{}_{}'.format(basename, str(i_node).zfill(6)))
        ctx.obj['NAME'] = name
        node_config['job_name'] = name

        cmd = BiasCalcClass.get_node_cmd(node_config)

        cmd_log = '\n\t'.join(cmd.split('\n'))
        logger.debug(f'Running command:\n\t{cmd_log}')

        if hardware_option.lower() in ('eagle', 'slurm'):
            kickoff_slurm_job(ctx, cmd, **exec_kwargs)
        else:
            kickoff_local_job(ctx, cmd)


def kickoff_slurm_job(ctx, cmd, alloc='sup3r', memory=None, walltime=4,
                      feature=None, stdout_path='./stdout/'):
    """Run sup3r on HPC via SLURM job submission.

    Parameters
    ----------
    ctx : click.pass_context
        Click context object where ctx.obj is a dictionary
    cmd : str
        Command to be submitted in SLURM shell script. Example:
            'python -m sup3r.cli forward_pass -c <config_file>'
    alloc : str
        HPC project (allocation) handle. Example: 'sup3r'.
    memory : int
        Node memory request in GB.
    walltime : float
        Node walltime request in hours.
    feature : str
        Additional flags for SLURM job. Format is "--qos=high"
        or "--depend=[state:job_id]". Default is None.
    stdout_path : str
        Path to print .stdout and .stderr files.
    """
    BaseCLI.kickoff_slurm_job(ModuleName.BIAS_CALC, ctx, cmd, alloc, memory,
                              walltime, feature, stdout_path)


def kickoff_local_job(ctx, cmd):
    """Run sup3r bias calc locally.

    Parameters
    ----------
    ctx : click.pass_context
        Click context object where ctx.obj is a dictionary
    cmd : str
        Command to be submitted in shell script. Example:
            'python -m sup3r.cli forward_pass -c <config_file>'
    """
    BaseCLI.kickoff_local_job(ModuleName.BIAS_CALC, ctx, cmd)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running sup3r bias calc CLI')
        raise



# -*- coding: utf-8 -*-
"""
Sup3r batch_handling module.
"""
import logging
import numpy as np
from datetime import datetime as dt

from rex.utilities import log_mem

from sup3r.utilities.utilities import (spatial_coarsening,
                                       temporal_coarsening,
                                       spatial_simple_enhancing,
                                       temporal_simple_enhancing,
                                       smooth_data)
from sup3r.preprocessing.batch_handling import (Batch,
                                                ValidationData,
                                                BatchHandler)

np.random.seed(42)

logger = logging.getLogger(__name__)


class BatchMom1(Batch):
    """Batch of low_res, high_res and output data"""

    def __init__(self, low_res, high_res, output, mask):
        """Stores low, high res, output and mask data

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        output : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        mask : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        """
        self._low_res = low_res
        self._high_res = high_res
        self._output = output
        self._mask = mask

    @property
    def output(self):
        """Get the output for the batch.
           Output predicted by the neural net can be different
           than the high_res when doing moment estimation.
           For ex: output may be (high_res)**2
           We distinguish output from high_res since it may not be
           possible to recover high_res from output."""
        return self._output

    @property
    def mask(self):
        """Get the mask for the batch."""
        return self._mask

    # pylint: disable=W0613
    @staticmethod
    def make_output(low_res, high_res,
                    s_enhance=None, t_enhance=None,
                    model_mom1=None, output_features_ind=None):
        """Make custom batch output

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int | None
            Spatial enhancement factor
        t_enhance : int | None
            Temporal enhancement factor
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        return high_res

    # pylint: disable=E1130
    @staticmethod
    def make_mask(high_res, s_padding=None, t_padding=None):
        """Make mask for output.
        The mask is used to ensure consistency when training conditional
        moments.
        Consider the case of learning E(HR|LR) where HR is the high_res and
        LR is the low_res.
        In theory, the conditional moment estimation works if
        the full LR is passed as input and predicts the full HR.
        In practice, only the LR data that overlaps and surrounds the HR data
        is useful, ie E(HR|LR) = E(HR|LR_nei) where LR_nei is the LR data
        that surrounds the HR data. Physically, this is equivalent to saying
        that data far away from a region of interest does not matter.
        This allows learning the conditional moments on spatial and
        temporal chunks only if one restricts the high_res output as being
        overlapped and surrounded by the input low_res.
        The role of the mask is to ensure that the input low_res always
        surrounds the output high_res.

        Parameters
        ----------
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_padding : int | None
            Spatial padding size. If None or 0, no padding is applied.
            None by default
        t_padding : int | None
            Temporal padding size. If None or 0, no padding is applied.
            None by default
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        mask = np.zeros(high_res.shape, dtype=np.float32)
        s_min = s_padding if s_padding is not None else 0
        t_min = t_padding if t_padding is not None else 0
        s_max = -s_padding if s_min > 0 else None
        t_max = -t_padding if t_min > 0 else None

        if len(high_res.shape) == 4:
            mask[:, s_min:s_max, s_min:s_max, :] = 1.0
        elif len(high_res.shape) == 5:
            mask[:, s_min:s_max, s_min:s_max, t_min:t_max, :] = 1.0

        return mask

    # pylint: disable=W0613
    @classmethod
    def get_coarse_batch(cls, high_res,
                         s_enhance, t_enhance=1,
                         temporal_coarsening_method='subsample',
                         output_features_ind=None,
                         output_features=None,
                         training_features=None,
                         smoothing=None,
                         smoothing_ignore=None,
                         model_mom1=None,
                         s_padding=None,
                         t_padding=None):
        """Coarsen high res data and return Batch with high res and
        low res data

        Parameters
        ----------
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int
            Factor by which to coarsen spatial dimensions of the high
            resolution data
        t_enhance : int
            Factor by which to coarsen temporal dimension of the high
            resolution data
        temporal_coarsening_method : str
            Method to use for temporal coarsening. Can be subsample, average,
            or total
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        output_features : list
            List of Generative model output feature names
        training_features : list | None
            Ordered list of training features input to the generative model
        smoothing : float | None
            Standard deviation to use for gaussian filtering of the coarse
            data. This can be tuned by matching the kinetic energy of a low
            resolution simulation with the kinetic energy of a coarsened and
            smoothed high resolution simulation. If None no smoothing is
            performed.
        smoothing_ignore : list | None
            List of features to ignore for the smoothing filter. None will
            smooth all features if smoothing kwarg is not None
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        s_padding : int | None
            Width of spatial padding to predict only middle part. If None,
            no padding is used
        t_padding : int | None
            Width of temporal padding to predict only middle part. If None,
            no padding is used

        Returns
        -------
        Batch
            Batch instance with low and high res data
        """
        low_res = spatial_coarsening(high_res, s_enhance)

        if training_features is None:
            training_features = [None] * low_res.shape[-1]

        if smoothing_ignore is None:
            smoothing_ignore = []

        if t_enhance != 1:
            low_res = temporal_coarsening(low_res, t_enhance,
                                          temporal_coarsening_method)

        low_res = smooth_data(low_res, training_features, smoothing_ignore,
                              smoothing)
        high_res = cls.reduce_features(high_res, output_features_ind)
        output = cls.make_output(low_res, high_res,
                                 s_enhance, t_enhance,
                                 model_mom1, output_features_ind)
        mask = cls.make_mask(high_res,
                             s_padding, t_padding)
        batch = cls(low_res, high_res, output, mask)

        return batch


class BatchMom1SF(BatchMom1):
    """Batch of low_res, high_res and output data
    when learning first moment of subfilter vel"""

    @staticmethod
    def make_output(low_res, high_res,
                    s_enhance=None, t_enhance=None,
                    model_mom1=None, output_features_ind=None):
        """Make custom batch output

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int | None
            Spatial enhancement factor
        t_enhance : int | None
            Temporal enhancement factor
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        # Remove LR from HR
        enhanced_lr = spatial_simple_enhancing(low_res,
                                               s_enhance=s_enhance)
        enhanced_lr = temporal_simple_enhancing(enhanced_lr,
                                                t_enhance=t_enhance)
        enhanced_lr = Batch.reduce_features(enhanced_lr, output_features_ind)
        return high_res - enhanced_lr


class BatchMom2(BatchMom1):
    """Batch of low_res, high_res and output data
    when learning second moment"""

    @staticmethod
    def make_output(low_res, high_res,
                    s_enhance=None, t_enhance=None,
                    model_mom1=None, output_features_ind=None):
        """Make custom batch output

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int | None
            Spatial enhancement factor
        t_enhance : int | None
            Temporal enhancement factor
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        # Remove first moment from HR and square it
        out = model_mom1._tf_generate(low_res).numpy()
        return (high_res - out)**2


class BatchMom2Sep(BatchMom1):
    """Batch of low_res, high_res and output data
    when learning second moment separate from first moment"""

    @staticmethod
    def make_output(low_res, high_res,
                    s_enhance=None, t_enhance=None,
                    model_mom1=None, output_features_ind=None):
        """Make custom batch output

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int | None
            Spatial enhancement factor
        t_enhance : int | None
            Temporal enhancement factor
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        return super(BatchMom2Sep,
                     BatchMom2Sep).make_output(low_res, high_res,
                                               s_enhance, t_enhance,
                                               model_mom1,
                                               output_features_ind)**2


class BatchMom2SF(BatchMom1):
    """Batch of low_res, high_res and output data
    when learning second moment of subfilter vel"""

    @staticmethod
    def make_output(low_res, high_res,
                    s_enhance=None, t_enhance=None,
                    model_mom1=None, output_features_ind=None):
        """Make custom batch output

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int | None
            Spatial enhancement factor
        t_enhance : int | None
            Temporal enhancement factor
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        # Remove LR and first moment from HR and square it
        out = model_mom1._tf_generate(low_res).numpy()
        enhanced_lr = spatial_simple_enhancing(low_res,
                                               s_enhance=s_enhance)
        enhanced_lr = temporal_simple_enhancing(enhanced_lr,
                                                t_enhance=t_enhance)
        enhanced_lr = Batch.reduce_features(enhanced_lr, output_features_ind)
        return (high_res - enhanced_lr - out)**2


class BatchMom2SepSF(BatchMom1SF):
    """Batch of low_res, high_res and output data
    when learning second moment of subfilter vel
    separate from first moment"""

    @staticmethod
    def make_output(low_res, high_res,
                    s_enhance=None, t_enhance=None,
                    model_mom1=None, output_features_ind=None):
        """Make custom batch output

        Parameters
        ----------
        low_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)
        s_enhance : int | None
            Spatial enhancement factor
        t_enhance : int | None
            Temporal enhancement factor
        model_mom1 : Sup3rCondMom | None
            Model used to modify the make the batch output
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        """
        # Remove LR from HR and square it
        return super(BatchMom2SepSF,
                     BatchMom2SepSF).make_output(low_res, high_res,
                                                 s_enhance, t_enhance,
                                                 model_mom1,
                                                 output_features_ind)**2


class ValidationDataMom1(ValidationData):
    """Iterator for validation data"""

    # Classes to use for handling an individual batch obj.
    BATCH_CLASS = BatchMom1

    def __init__(self, data_handlers, batch_size=8, s_enhance=3, t_enhance=1,
                 temporal_coarsening_method='subsample',
                 output_features_ind=None,
                 output_features=None,
                 smoothing=None, smoothing_ignore=None,
                 model_mom1=None,
                 s_padding=None, t_padding=None):
        """
        Parameters
        ----------
        handlers : list[DataHandler]
            List of DataHandler instances
        batch_size : int
            Size of validation data batches
        s_enhance : int
            Factor by which to coarsen spatial dimensions of the high
            resolution data
        t_enhance : int
            Factor by which to coarsen temporal dimension of the high
            resolution data
        temporal_coarsening_method : str
            [subsample, average, total]
            Subsample will take every t_enhance-th time step, average will
            average over t_enhance time steps, total will sum over t_enhance
            time steps
        output_features_ind : list | np.ndarray | None
            List/array of feature channel indices that are used for generative
            output, without any feature indices used only for training.
        output_features : list
            List of Generative model output feature names
        smoothing : float | None
            Standard deviation to use for gaussian filtering of the coarse
            data. This can be tuned by matching the kinetic energy of a low
            resolution simulation with the kinetic energy of a coarsened and
            smoothed high resolution simulation. If None no smoothing is
            performed.
        smoothing_ignore : list | None
            List of features to ignore for the smoothing filter. None will
            smooth all features if smoothing kwarg is not None
        model_mom1 : Sup3rCondMom | None
            model that predicts the first conditional moments.
            Useful to prepare data for learning second conditional moment.
        s_padding : int | None
            Width of spatial padding to predict only middle part. If None,
            no padding is used
        t_padding : int | None
            Width of temporal padding to predict only middle part. If None,
            no padding is used
        """

        handler_shapes = np.array([d.sample_shape for d in data_handlers])
        assert np.all(handler_shapes[0] == handler_shapes)

        self.handlers = data_handlers
        self.batch_size = batch_size
        self.sample_shape = handler_shapes[0]
        self.val_indices = self._get_val_indices()
        self.max = np.ceil(
            len(self.val_indices) / (batch_size))
        self.s_enhance = s_enhance
        self.t_enhance = t_enhance
        self.s_padding = s_padding
        self.t_padding = t_padding
        self._remaining_observations = len(self.val_indices)
        self.temporal_coarsening_method = temporal_coarsening_method
        self._i = 0
        self.output_features_ind = output_features_ind
        self.output_features = output_features
        self.smoothing = smoothing
        self.smoothing_ignore = smoothing_ignore
        self.model_mom1 = model_mom1

    def batch_next(self, high_res):
        """Assemble the next batch

        Parameters
        ----------
        high_res : np.ndarray
            4D | 5D array
            (batch_size, spatial_1, spatial_2, features)
            (batch_size, spatial_1, spatial_2, temporal, features)

        Returns
        -------
        batch : Batch
        """
        return self.BATCH_CLASS.get_coarse_batch(
            high_res, self.s_enhance,
            t_enhance=self.t_enhance,
            temporal_coarsening_method=self.temporal_coarsening_method,
            output_features_ind=self.output_features_ind,
            smoothing=self.smoothing,
            smoothing_ignore=self.smoothing_ignore,
            output_features=self.output_features,
            model_mom1=self.model_mom1,
            s_padding=self.s_padding,
            t_padding=self.t_padding)


class BatchHandlerMom1(BatchHandler):
    """Sup3r base batch handling class"""

    # Classes to use for handling an individual batch obj.
    VAL_CLASS = ValidationDataMom1
    BATCH_CLASS = BatchMom1
    DATA_HANDLER_CLASS = None

    def __init__(self, data_handlers, batch_size=8, s_enhance=3, t_enhance=1,
                 means=None, stds=None, norm=True, n_batches=10,
                 temporal_coarsening_method='subsample', stdevs_file=None,
                 means_file=None, overwrite_stats=False, smoothing=None,
                 smoothing_ignore=None, stats_workers=None, norm_workers=None,
                 load_workers=None, max_workers=None, model_mom1=None,
                 s_padding=None, t_padding=None):
        """
        Parameters
        ----------
        data_handlers : list[DataHandler]
            List of DataHandler instances
        batch_size : int
            Number of observations in a batch
        s_enhance : int
            Factor by which to coarsen spatial dimensions of the high
            resolution data to generate low res data
        t_enhance : int
            Factor by which to coarsen temporal dimension of the high
            resolution data to generate low res data
        means : np.ndarray
            dimensions (features)
            array of means for all features with same ordering as data
            features. If not None and norm is True these will be used for
            normalization
        stds : np.ndarray
            dimensions (features)
            array of means for all features with same ordering as data
            features.  If not None and norm is True these will be used form
            normalization
        norm : bool
            Whether to normalize the data or not
        n_batches : int
            Number of batches in an epoch, this sets the iteration limit for
            this object.
        temporal_coarsening_method : str
            [subsample, average, total]
            Subsample will take every t_enhance-th time step, average will
            average over t_enhance time steps, total will sum over t_enhance
            time steps
        stdevs_file : str | None
            Path to stdevs data or where to save data after calling get_stats
        means_file : str | None
            Path to means data or where to save data after calling get_stats
        overwrite_stats : bool
            Whether to overwrite stats cache files.
        smoothing : float | None
            Standard deviation to use for gaussian filtering of the coarse
            data. This can be tuned by matching the kinetic energy of a low
            resolution simulation with the kinetic energy of a coarsened and
            smoothed high resolution simulation. If None no smoothing is
            performed.
        smoothing_ignore : list | None
            List of features to ignore for the smoothing filter. None will
            smooth all features if smoothing kwarg is not None
        max_workers : int | None
            Providing a value for max workers will be used to set the value of
            norm_workers, stats_workers, and load_workers.
            If max_workers == 1 then all processes will be serialized. If None
            stats_workers, load_workers, and norm_workers will use their own
            provided values.
        load_workers : int | None
            max number of workers to use for loading data handlers.
        norm_workers : int | None
            max number of workers to use for normalizing data handlers.
        stats_workers : int | None
            max number of workers to use for computing stats across data
            handlers.
        model_mom1 : Sup3rCondMom | None
            model that predicts the first conditional moments.
            Useful to prepare data for learning second conditional moment.
        s_padding : int | None
            Width of spatial padding to predict only middle part. If None,
            no padding is used
        t_padding : int | None
            Width of temporal padding to predict only middle part. If None,
            no padding is used
        """
        if max_workers is not None:
            norm_workers = stats_workers = load_workers = max_workers

        msg = ('All data handlers must have the same sample_shape')
        handler_shapes = np.array([d.sample_shape for d in data_handlers])
        assert np.all(handler_shapes[0] == handler_shapes), msg

        self.data_handlers = data_handlers
        self._i = 0
        self.low_res = None
        self.high_res = None
        self.output = None
        self.batch_size = batch_size
        self._val_data = None
        self.s_enhance = s_enhance
        self.t_enhance = t_enhance
        self.s_padding = s_padding
        self.t_padding = t_padding
        self.sample_shape = handler_shapes[0]
        self.means = means
        self.stds = stds
        self.n_batches = n_batches
        self.temporal_coarsening_method = temporal_coarsening_method
        self.current_batch_indices = None
        self.current_handler_index = None
        self.stdevs_file = stdevs_file
        self.means_file = means_file
        self.overwrite_stats = overwrite_stats
        self.smoothing = smoothing
        self.smoothing_ignore = smoothing_ignore or []
        self.smoothed_features = [f for f in self.training_features
                                  if f not in self.smoothing_ignore]
        self._stats_workers = stats_workers
        self._norm_workers = norm_workers
        self._load_workers = load_workers
        self.model_mom1 = model_mom1

        logger.info(f'Initializing BatchHandler with smoothing={smoothing}. '
                    f'Using stats_workers={self.stats_workers}, '
                    f'norm_workers={self.norm_workers}, '
                    f'load_workers={self.load_workers}.')

        now = dt.now()
        self.parallel_load()
        logger.debug(f'Finished loading data of shape {self.shape} '
                     f'for BatchHandler in {dt.now() - now}.')
        log_mem(logger, log_level='INFO')

        if norm:
            self.means, self.stds = self.check_cached_stats()
            self.normalize(self.means, self.stds)

        logger.debug('Getting validation data for BatchHandler.')
        self.val_data = self.VAL_CLASS(
            data_handlers, batch_size=batch_size,
            s_enhance=s_enhance, t_enhance=t_enhance,
            temporal_coarsening_method=temporal_coarsening_method,
            output_features_ind=self.output_features_ind,
            output_features=self.output_features,
            smoothing=self.smoothing,
            smoothing_ignore=self.smoothing_ignore,
            model_mom1=self.model_mom1,
            s_padding=self.s_padding,
            t_padding=self.t_padding)

        logger.info('Finished initializing BatchHandler.')
        log_mem(logger, log_level='INFO')

    def __next__(self):
        """Get the next iterator output.

        Returns
        -------
        batch : Batch
            Batch object with batch.low_res, batch.high_res
            and batch.output attributes with the appropriate coarsening.
        """
        self.current_batch_indices = []
        if self._i < self.n_batches:
            handler_index = np.random.randint(0, len(self.data_handlers))
            self.current_handler_index = handler_index
            handler = self.data_handlers[handler_index]
            high_res = np.zeros((self.batch_size, self.sample_shape[0],
                                 self.sample_shape[1], self.sample_shape[2],
                                 self.shape[-1]), dtype=np.float32)

            for i in range(self.batch_size):
                high_res[i, ...] = handler.get_next()
                self.current_batch_indices.append(handler.current_obs_index)

            batch = self.BATCH_CLASS.get_coarse_batch(
                high_res, self.s_enhance, t_enhance=self.t_enhance,
                temporal_coarsening_method=self.temporal_coarsening_method,
                output_features_ind=self.output_features_ind,
                output_features=self.output_features,
                training_features=self.training_features,
                smoothing=self.smoothing,
                smoothing_ignore=self.smoothing_ignore,
                model_mom1=self.model_mom1,
                s_padding=self.s_padding,
                t_padding=self.t_padding)

            self._i += 1
            return batch
        else:
            raise StopIteration


class SpatialBatchHandlerMom1(BatchHandlerMom1):
    """Sup3r spatial batch handling class"""

    def __next__(self):
        if self._i < self.n_batches:
            handler_index = np.random.randint(
                0, len(self.data_handlers))
            handler = self.data_handlers[handler_index]
            high_res = np.zeros((self.batch_size, self.sample_shape[0],
                                 self.sample_shape[1], self.shape[-1]),
                                dtype=np.float32)
            for i in range(self.batch_size):
                high_res[i, ...] = handler.get_next()[..., 0, :]

            batch = self.BATCH_CLASS.get_coarse_batch(
                high_res, self.s_enhance,
                output_features_ind=self.output_features_ind,
                training_features=self.training_features,
                smoothing=self.smoothing,
                smoothing_ignore=self.smoothing_ignore,
                model_mom1=self.model_mom1,
                s_padding=self.s_padding,
                t_padding=self.t_padding)

            self._i += 1
            return batch
        else:
            raise StopIteration


class ValidationDataMom1SF(ValidationDataMom1):
    """Iterator for validation data for
    first conditional moment of subfilter velocity"""
    BATCH_CLASS = BatchMom1SF


class ValidationDataMom2(ValidationDataMom1):
    """Iterator for subfilter validation data for
    second conditional moment"""
    BATCH_CLASS = BatchMom2


class ValidationDataMom2Sep(ValidationDataMom1):
    """Iterator for subfilter validation data for
    second conditional moment separate from first
    moment"""
    BATCH_CLASS = BatchMom2Sep


class ValidationDataMom2SF(ValidationDataMom1):
    """Iterator for validation data for
    second conditional moment of subfilter velocity"""
    BATCH_CLASS = BatchMom2SF


class ValidationDataMom2SepSF(ValidationDataMom1):
    """Iterator for validation data for
    second conditional moment of subfilter velocity
    separate from first moment"""
    BATCH_CLASS = BatchMom2SepSF


class BatchHandlerMom1SF(BatchHandlerMom1):
    """Sup3r batch handling class for
    first conditional moment of subfilter velocity"""
    VAL_CLASS = ValidationDataMom1SF
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class SpatialBatchHandlerMom1SF(SpatialBatchHandlerMom1):
    """Sup3r spatial batch handling class for
    first conditional moment of subfilter velocity"""
    VAL_CLASS = ValidationDataMom1SF
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class BatchHandlerMom2(BatchHandlerMom1):
    """Sup3r batch handling class for
    second conditional moment"""
    VAL_CLASS = ValidationDataMom2
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class BatchHandlerMom2Sep(BatchHandlerMom1):
    """Sup3r batch handling class for
    second conditional moment separate from first
    moment"""
    VAL_CLASS = ValidationDataMom2Sep
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class SpatialBatchHandlerMom2(SpatialBatchHandlerMom1):
    """Sup3r spatial batch handling class for
    second conditional moment"""
    VAL_CLASS = ValidationDataMom2
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class SpatialBatchHandlerMom2Sep(SpatialBatchHandlerMom1):
    """Sup3r spatial batch handling class for
    second conditional moment separate from first
    moment"""
    VAL_CLASS = ValidationDataMom2Sep
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class BatchHandlerMom2SF(BatchHandlerMom1):
    """Sup3r batch handling class for
    second conditional moment of subfilter velocity"""
    VAL_CLASS = ValidationDataMom2SF
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class BatchHandlerMom2SepSF(BatchHandlerMom1):
    """Sup3r batch handling class for
    second conditional moment of subfilter velocity
    separate from first moment"""
    VAL_CLASS = ValidationDataMom2SepSF
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class SpatialBatchHandlerMom2SF(SpatialBatchHandlerMom1):
    """Sup3r spatial batch handling class for
    second conditional moment of subfilter velocity"""
    VAL_CLASS = ValidationDataMom2SF
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS


class SpatialBatchHandlerMom2SepSF(SpatialBatchHandlerMom1):
    """Sup3r spatial batch handling class for
    second conditional moment of subfilter velocity
    separate from first moment"""
    VAL_CLASS = ValidationDataMom2SepSF
    BATCH_CLASS = VAL_CLASS.BATCH_CLASS



# -*- coding: utf-8 -*-
"""Execution methods for running some cli routines

@author: bbenton
"""

import numpy as np
import logging
import os

logger = logging.getLogger(__name__)


class DistributedProcess:
    """High-level class with commonly used functionality for processes
    distributed across multiple nodes"""

    def __init__(self, max_nodes=1, n_chunks=None, max_chunks=None,
                 incremental=False):
        """
        Parameters
        ----------
        max_nodes : int, optional
            Max number of nodes to distribute processes across
        n_chunks : int, optional
            Number of chunks to split all processes into. These process
            chunks will be distributed across nodes.
        max_chunks : int, optional
            Max number of chunks processes can be split into.
        incremental : bool
            Whether to skip previously run process chunks or to overwrite.
        """
        msg = ('For a distributed process either max_chunks or '
               'max_chunks + n_chunks must be specified. Received '
               f'max_chunks={max_chunks}, n_chunks={n_chunks}.')
        assert max_chunks is not None, msg
        self._node_chunks = None
        self._n_chunks = n_chunks
        self._max_nodes = max_nodes
        self._max_chunks = max_chunks
        self._out_files = None
        self._failed_chunks = False
        self.incremental = incremental

    def __len__(self):
        """Get total number of process chunks"""
        return self.chunks

    def node_finished(self, node_index):
        """Check if all out files for a given node have been saved

        Parameters
        ----------
        node_index : int
            Index of node to check for completed processes

        Returns
        -------
        bool
            Whether all processes for the given node have finished
        """
        return all(self.chunk_finished(i)
                   for i in self.node_chunks[node_index])

    # pylint: disable=E1136
    def chunk_finished(self, chunk_index):
        """Check if process for given chunk_index has already been run.

        Parameters
        ----------
        chunk_index : int
            Index of the process chunk to check for completion. Considered
            finished if there is already an output file and incremental is
            False.

        Returns
        -------
        bool
            Whether the process for the given chunk has finished
        """
        out_file = self.out_files[chunk_index]
        if os.path.exists(out_file) and self.incremental:
            logger.info('Not running chunk index {}, output file '
                        'exists: {}'.format(chunk_index, out_file))
            return True
        return False

    @property
    def all_finished(self):
        """Check if all out files have been saved"""
        return all(self.node_finished(i) for i in range(self.nodes))

    @property
    def out_files(self):
        """Get list of out files to write process output to"""
        return self._out_files

    @property
    def max_nodes(self):
        """Get uncapped max number of nodes to distribute processes across"""
        return self._max_nodes

    @property
    def chunks(self):
        """Get the number of process chunks for this distributed routine."""
        if self._n_chunks is None:
            return self._max_chunks
        else:
            return min(self._n_chunks, self._max_chunks)

    @property
    def nodes(self):
        """Get the max number of nodes to distribute chunks across, limited by
        the number of process chunks"""
        return len(self.node_chunks)

    @property
    def node_chunks(self):
        """Get the chunk indices for different nodes"""
        if self._node_chunks is None:
            n_chunks = min(self.max_nodes, self.chunks)
            self._node_chunks = np.array_split(np.arange(self.chunks),
                                               n_chunks)
        return self._node_chunks

    @property
    def failed_chunks(self):
        """Check whether any processes have failed."""
        return self._failed_chunks

    @failed_chunks.setter
    def failed_chunks(self, failed):
        """Set failed_chunks value. Should be set to True if there is a failed
        chunk"""
        self._failed_chunks = failed



# -*- coding: utf-8 -*-
"""Bias correction transformation functions."""
import numpy as np
import logging
from warnings import warn
from rex import Resource


logger = logging.getLogger(__name__)


def global_linear_bc(input, scalar, adder, out_range=None):
    """Bias correct data using a simple global *scalar +adder method.

    Parameters
    ----------
    input : np.ndarray
        Sup3r input data to be bias corrected, assumed to be 3D with shape
        (spatial, spatial, temporal) for a single feature.
    scalar : float
        Scalar (multiplicative) value to apply to input data.
    adder : float
        Adder value to apply to input data.
    out_range : None | tuple
        Option to set floor/ceiling values on the output data.

    Returns
    -------
    out : np.ndarray
        out = input * scalar + adder
    """
    out = input * scalar + adder
    if out_range is not None:
        out = np.maximum(out, np.min(out_range))
        out = np.minimum(out, np.max(out_range))
    return out


def local_linear_bc(input, feature_name, bias_fp, lr_padded_slice,
                    out_range=None):
    """Bias correct data using a simple annual (or multi-year) *scalar +adder
    method on a site-by-site basis.

    Parameters
    ----------
    input : np.ndarray
        Sup3r input data to be bias corrected, assumed to be 3D with shape
        (spatial, spatial, temporal) for a single feature.
    feature_name : str
        Name of feature that is being corrected. Datasets with names
        "{feature_name}_scalar" and "{feature_name}_adder" will be retrieved
        from bias_fp.
    bias_fp : str
        Filepath to bias correction file from the bias calc module. Must have
        datasets "{feature_name}_scalar" and "{feature_name}_adder" that are
        the full low-resolution shape of the forward pass input that will be
        sliced using lr_padded_slice for the current chunk.
    lr_padded_slice : tuple | None
        Tuple of length four that slices (spatial_1, spatial_2, temporal,
        features) where each tuple entry is a slice object for that axes.
        Note that if this method is called as part of a sup3r forward pass, the
        lr_padded_slice will be included in the kwargs for the active chunk.
        If this is None, no slicing will be done and the full bias correction
        source shape will be used.
    out_range : None | tuple
        Option to set floor/ceiling values on the output data.

    Returns
    -------
    out : np.ndarray
        out = input * scalar + adder
    """

    scalar = f'{feature_name}_scalar'
    adder = f'{feature_name}_adder'
    with Resource(bias_fp) as res:
        scalar = res[scalar]
        adder = res[adder]

    # 3D bias correction factors have seasonal/monthly correction in last axis
    if len(scalar.shape) == 3 and len(adder.shape) == 3:
        scalar = scalar.mean(axis=-1)
        adder = adder.mean(axis=-1)

    if lr_padded_slice is not None:
        spatial_slice = (lr_padded_slice[0], lr_padded_slice[1])
        scalar = scalar[spatial_slice]
        adder = adder[spatial_slice]

    if np.isnan(scalar).any() or np.isnan(adder).any():
        msg = ('Bias correction scalar/adder values had NaNs for "{}" from: {}'
               .format(feature_name, bias_fp))
        logger.warning(msg)
        warn(msg)

    scalar = np.expand_dims(scalar, axis=-1)
    adder = np.expand_dims(adder, axis=-1)

    scalar = np.repeat(scalar, input.shape[-1], axis=-1)
    adder = np.repeat(adder, input.shape[-1], axis=-1)

    out = input * scalar + adder
    if out_range is not None:
        out = np.maximum(out, np.min(out_range))
        out = np.minimum(out, np.max(out_range))

    return out


def monthly_local_linear_bc(input, feature_name, bias_fp, lr_padded_slice,
                            time_index, temporal_avg=True, out_range=None):
    """Bias correct data using a simple monthly *scalar +adder method on a
    site-by-site basis.

    Parameters
    ----------
    input : np.ndarray
        Sup3r input data to be bias corrected, assumed to be 3D with shape
        (spatial, spatial, temporal) for a single feature.
    feature_name : str
        Name of feature that is being corrected. Datasets with names
        "{feature_name}_scalar" and "{feature_name}_adder" will be retrieved
        from bias_fp.
    bias_fp : str
        Filepath to bias correction file from the bias calc module. Must have
        datasets "{feature_name}_scalar" and "{feature_name}_adder" that are
        the full low-resolution shape of the forward pass input that will be
        sliced using lr_padded_slice for the current chunk.
    lr_padded_slice : tuple | None
        Tuple of length four that slices (spatial_1, spatial_2, temporal,
        features) where each tuple entry is a slice object for that axes.
        Note that if this method is called as part of a sup3r forward pass, the
        lr_padded_slice will be included automatically in the kwargs for the
        active chunk. If this is None, no slicing will be done and the full
        bias correction source shape will be used.
    time_index : pd.DatetimeIndex
        DatetimeIndex object associated with the input data temporal axis
        (assumed 3rd axis e.g. axis=2). Note that if this method is called as
        part of a sup3r resolution forward pass, the time_index will be
        included automatically for the current chunk.
    temporal_avg : bool
        Take the average scalars and adders for the chunk's time index, this
        will smooth the transition of scalars/adders from month to month if
        processing small chunks. If processing the full annual time index, set
        this to False.
    out_range : None | tuple
        Option to set floor/ceiling values on the output data.

    Returns
    -------
    out : np.ndarray
        out = input * scalar + adder
    """

    scalar = f'{feature_name}_scalar'
    adder = f'{feature_name}_adder'
    with Resource(bias_fp) as res:
        scalar = res[scalar]
        adder = res[adder]

    assert len(scalar.shape) == 3, 'Monthly bias correct needs 3D scalars'
    assert len(adder.shape) == 3, 'Monthly bias correct needs 3D adders'

    if lr_padded_slice is not None:
        spatial_slice = (lr_padded_slice[0], lr_padded_slice[1])
        scalar = scalar[spatial_slice]
        adder = adder[spatial_slice]

    imonths = time_index.month.values - 1
    scalar = scalar[..., imonths]
    adder = adder[..., imonths]

    if temporal_avg:
        scalar = scalar.mean(axis=-1)
        adder = adder.mean(axis=-1)
        scalar = np.expand_dims(scalar, axis=-1)
        adder = np.expand_dims(adder, axis=-1)
        scalar = np.repeat(scalar, input.shape[-1], axis=-1)
        adder = np.repeat(adder, input.shape[-1], axis=-1)
        if len(time_index.month.unique()) > 2:
            msg = ('Bias correction method "monthly_local_linear_bc" was used '
                   'with temporal averaging over a time index with >2 months.')
            warn(msg)
            logger.warning(msg)

    if np.isnan(scalar).any() or np.isnan(adder).any():
        msg = ('Bias correction scalar/adder values had NaNs for "{}" from: {}'
               .format(feature_name, bias_fp))
        logger.warning(msg)
        warn(msg)

    out = input * scalar + adder
    if out_range is not None:
        out = np.maximum(out, np.min(out_range))
        out = np.minimum(out, np.max(out_range))

    return out



"""Utilities for stitching south east asia domains

Example Use:

    stitch_and_save(year=2017, month=1,
                    input_pattern="wrfout_d0{domain}_{year}-{month}*",
                    out_pattern="{year}/{month}/", overlap=15, n_domains=4,
                    max_levels=10)

    This will combine 4 domains, for Jan 2017, using an overlap of 15 grid
    points the blend the domain edges, and save only the first 10 pressure
    levels. The stitched files will be saved in the directory specified with
    out_pattern.
"""
# -*- coding: utf-8 -*-
import xarray as xr
import numpy as np
import logging
from importlib import import_module
import glob
import os

logger = logging.getLogger(__name__)

SAVE_FEATURES = ['U', 'V', 'PHB', 'PH', 'HGT', 'P', 'PB', 'T', 'Times']


class Regridder:
    """Regridder class for stitching domains"""

    DEPENDENCIES = ['xesmf']

    def __init__(self, lats, lons, min_lat, max_lat, min_lon, max_lon,
                 n_lats, n_lons):
        """
        Parameters
        ----------
        lats : ndarray
            Array of latitudes for input grid
        lons : ndarray
            Array of longitudes for input grid
        min_lat : float
            Minimum lat for output grid
        max_lat : float
            Maximum lat for output grid
        min_lon : float
            Minimum lon for output grid
        max_lon : float
            Maximum lon for output grid
        n_lats : int
            Number of lats for output grid
        n_lons : int
            Number of lons for output grid
        """
        self.check_dependencies()
        import xesmf as xe
        self.grid_in = {'lat': lats, 'lon': lons}
        lons, lats = np.meshgrid(np.linspace(min_lon, max_lon, n_lons),
                                 np.linspace(min_lat, max_lat, n_lats))
        self.grid_out = {'lat': lats, 'lon': lons}
        self.new_lat_lon = np.zeros((*lats.shape, 2))
        self.new_lat_lon[..., 0] = lats
        self.new_lat_lon[..., 1] = lons
        self.regridder = xe.Regridder(self.grid_in, self.grid_out,
                                      method='bilinear')

    @classmethod
    def check_dependencies(cls):
        """Check special dependencies for stitching module"""

        missing = []
        for name in cls.DEPENDENCIES:
            try:
                import_module(name)
            except ModuleNotFoundError:
                missing.append(name)

        if any(missing):
            msg = ('The sup3r stitching module depends on the following '
                   'special dependencies that were not found in the active '
                   'environment: {}'.format(missing))
            logger.error(msg)
            raise ModuleNotFoundError(msg)

    def regrid_data(self, data_in):
        """Regrid data to output grid

        Parameters
        ----------
        data_in : xarray.Dataset
            input data handle

        Returns
        -------
        data_out : xarray.Dataset
            output data handle
        """
        times = data_in.Times.values
        data_out = self.regridder(data_in)
        data_out = data_out.rename({'lat': 'XLAT', 'lon': 'XLONG'})
        data_out = data_out.rename({'x': 'west_east', 'y': 'south_north'})
        data_out['Times'] = ('Time', times)
        data_out['XLAT'] = (('Time', 'south_north', 'west_east'),
                            np.repeat(np.expand_dims(data_out['XLAT'].values,
                                                     axis=0),
                                      len(times), axis=0))
        data_out['XLONG'] = (('Time', 'south_north', 'west_east'),
                             np.repeat(np.expand_dims(data_out['XLONG'].values,
                                                      axis=0),
                                       len(times), axis=0))
        return data_out


def get_files(year, month, input_pattern, out_pattern, n_domains=4):
    """Get input files for all domains to stitch together, and output file
    name

    Parameters
    ----------
    year : int
        Year for input files
    month : int
        Month for input files
    input_pattern : str
        Pattern for input files. Assumes pattern contains {month}, {year}, and
        {domain}
    out_pattern : str
        Pattern for output files. Assumes pattern contains {month} and {year}
    n_domains : int
        Number of domains to stitch together

    Returns
    -------
    input_files : dict
        Dictionary of input files with keys corresponding to domain number
    out_files : list
        List of output file names for final stitched output
    """
    in_pattern = [input_pattern.format(year=year, month=str(month).zfill(2),
                                       domain=i)
                  for i in range(1, n_domains + 1)]
    input_files = {i: sorted(glob.glob(in_pattern[i]))
                   for i in range(n_domains)}
    out_pattern = out_pattern.format(year=year, month=str(month).zfill(2))
    out_files = [os.path.join(out_pattern,
                              os.path.basename(input_files[0][i]).replace(
                                  'custom_wrfout_d01', 'stitched_wrfout'))
                 for i in range(len(input_files[0]))]
    return input_files, out_files


def get_handles(input_files):
    """Get handles for all domains. Keep needed fields

    Parameters
    ----------
    input_files : list
        List of input files for each domain. First file needs to be the file
        for the largest domain.

    Returns
    -------
    handles : list
        List of xarray.Dataset objects for each domain
    """
    handles = []
    for f in input_files:
        logger.info(f'Getting handle for {f}')
        handle = xr.open_dataset(f)
        handle = handle[SAVE_FEATURES]
        handles.append(handle)
    return handles


def unstagger_vars(handles):
    """Unstagger variables for all handles

    Parameters
    ----------
    handles : list
        List of xarray.Dataset objects for each domain

    Returns
    -------
    handles : list
        List of xarray.Dataset objects for each domain, with unstaggered
        variables.
    """
    dims = ('Time', 'bottom_top', 'south_north', 'west_east')
    for i, handle in enumerate(handles):
        handles[i]['U'] = (dims, np.apply_along_axis(forward_avg, 3,
                                                     handle['U']))
        handles[i]['V'] = (dims, np.apply_along_axis(forward_avg, 2,
                                                     handle['V']))
        handles[i]['PHB'] = (dims, np.apply_along_axis(forward_avg, 1,
                                                       handle['PHB']))
        handles[i]['PH'] = (dims, np.apply_along_axis(forward_avg, 1,
                                                      handle['PH']))
    return handles


def prune_levels(handles, max_level=15):
    """Prune pressure levels to reduce memory footprint

    Parameters
    ----------
    handles : list
        List of xarray.Dataset objects for each domain
    max_level : int
        Max pressure level index

    Returns
    -------
    handles : list
        List of xarray.Dataset objects for each domain, with pruned pressure
        levels.
    """
    for i, handle in enumerate(handles):
        handles[i] = handle.loc[dict(bottom_top=slice(0, max_level))]
    return handles


def regrid_main_domain(handles):
    """Regrid largest domain

    Parameters
    ----------
    handles : list
        List of xarray.Dataset objects for each domain

    Returns
    -------
    handles : list
        List of xarray.Dataset objects for each domain, with unstaggered
        variables and pruned pressure levels.
    """
    min_lat = np.min(handles[0].XLAT)
    min_lon = np.min(handles[0].XLONG)
    max_lat = np.max(handles[0].XLAT)
    max_lon = np.max(handles[0].XLONG)
    n_lons = handles[0].XLAT.shape[-1]
    n_lats = handles[0].XLAT.shape[1]
    main_regridder = Regridder(handles[0].XLAT[0], handles[0].XLONG[0],
                               min_lat, max_lat, min_lon, max_lon,
                               3 * n_lats, 3 * n_lons)
    handles[0] = main_regridder.regrid_data(handles[0])
    return handles


def forward_avg(array_in):
    """Forward average for use in unstaggering"""
    return (array_in[:-1] + array_in[1:]) * 0.5


def blend_domains(arr1, arr2, overlap=50):
    """Blend smaller domain edges

    Parameters
    ----------
    arr1 : ndarray
        Data array for largest domain
    arr2 : ndarray
        Data array for nested domain to stitch into larger domain
    overlap : int
        Number of grid points to use for blending edges

    Returns
    -------
    out : ndarray
        Data array with smaller domain blended into larger domain
    """
    out = arr2.copy()
    for i in range(overlap):
        alpha = i / overlap
        beta = 1 - alpha
        out[..., i, :] = out[..., i, :] * alpha + arr1[..., i, :] * beta
        out[..., -i, :] = out[..., -i, :] * alpha + arr1[..., -i, :] * beta
        out[..., :, i] = out[..., :, i] * alpha + arr1[..., :, i] * beta
        out[..., :, -i] = out[..., :, -i] * alpha + arr1[..., :, -i] * beta
    return out


def get_domain_region(handles, domain_num):
    """Get range for smaller domain

    Parameters
    ----------
    handles : list
        List of xarray.Dataset objects for each domain
    domain_num : int
        Domain number to get grid range for

    Returns
    -------
    lat_range : slice
        Slice corresponding to lat range of smaller domain within larger domain
    lon_range : slice
        Slice corresponding to lon range of smaller domain within larger domain
    min_lat : float
        Minimum lat for smaller domain
    max_lat : float
        Maximum lat for smaller domain
    min_lon : float
        Minimum lon for smaller domain
    max_lon : float
        Maximum lon for smaller domain
    n_lats : int
        Number of lats for smaller domain
    n_lons : int
        Number of lons for smaller domain
    """
    lats = handles[0].XLAT[0, :, 0]
    lons = handles[0].XLONG[0, 0, :]
    min_lat = np.min(handles[domain_num].XLAT.values)
    min_lon = np.min(handles[domain_num].XLONG.values)
    max_lat = np.max(handles[domain_num].XLAT.values)
    max_lon = np.max(handles[domain_num].XLONG.values)
    lat_mask = (min_lat <= lats) & (lats <= max_lat)
    lon_mask = (min_lon <= lons) & (lons <= max_lon)
    lat_idx = np.arange(len(lats))
    lon_idx = np.arange(len(lons))
    lat_range = slice(lat_idx[lat_mask][0], lat_idx[lat_mask][-1] + 1)
    lon_range = slice(lon_idx[lon_mask][0], lon_idx[lon_mask][-1] + 1)
    n_lats = len(lat_idx[lat_mask])
    n_lons = len(lon_idx[lon_mask])
    return (lat_range, lon_range, min_lat, max_lat, min_lon, max_lon,
            n_lats, n_lons)


def impute_domain(handles, domain_num, overlap=50):
    """Impute smaller domain in largest domain

    Parameters
    ----------
    handles : list
        List of xarray.Dataset objects for each domain
    domain_num : int
        Domain number to stitch into largest domain
    overlap : int
        Number of grid points to use for blending edges

    Returns
    -------
    handles : list
        List of xarray.Dataset objects for each domain
    """
    out = get_domain_region(handles, domain_num)
    (lat_range, lon_range, min_lat, max_lat, min_lon,
     max_lon, n_lats, n_lons) = out
    regridder = Regridder(handles[domain_num].XLAT[0],
                          handles[domain_num].XLONG[0],
                          min_lat, max_lat, min_lon, max_lon, n_lats, n_lons)
    handles[domain_num] = regridder.regrid_data(handles[domain_num])
    for field in handles[0]:
        if field not in ['Times']:
            arr1 = handles[0][field].loc[dict(south_north=lat_range,
                                              west_east=lon_range)]
            arr2 = handles[domain_num][field]
            out = blend_domains(arr1, arr2, overlap=overlap)
            handles[0][field].loc[dict(south_north=lat_range,
                                       west_east=lon_range)] = out
    return handles


def stitch_domains(year, month, time_step, input_files, overlap=50,
                   n_domains=4, max_level=15):
    """Stitch all smaller domains into largest domain

    Parameters
    ----------
    year : int
        Year for input files
    month : int
        Month for input files
    time_step : int
        Time step for input files for the specified month. e.g. if year=2017,
        month=3, time_step=0 this will select the file for the first time step
        of 2017-03-01. If None then stitch and save will be done for full
        month.
    input_files : dict
        Dictionary of input files with keys corresponding to domain number
    overlap : int
        Number of grid points to use for blending edges
    n_domains : int
        Number of domains to stitch together
    max_level : int
        Max pressure level index

    Returns
    -------
    handles : list
        List of xarray.Dataset objects with smaller domains stitched into
        handles[0]
    """
    logger.info(f'Getting domain files for year={year}, month={month},'
                f' timestep={time_step}.')
    step_files = [input_files[d][time_step] for d in range(n_domains)]
    logger.info(f'Getting data handles for files: {step_files}')
    handles = get_handles(step_files)
    logger.info('Unstaggering variables for all handles')
    handles = unstagger_vars(handles)
    logger.info(f'Pruning pressure levels to level={max_level}')
    handles = prune_levels(handles, max_level=max_level)
    logger.info(f'Regridding main domain for year={year}, month={month}, '
                f'timestep={time_step}')
    handles = regrid_main_domain(handles)
    for j in range(1, n_domains):
        logger.info(f'Imputing domain {j + 1} for year={year}, '
                    f'month={month}, timestep={time_step}')
        handles = impute_domain(handles, j, overlap=overlap)
    return handles


def stitch_and_save(year, month, input_pattern, out_pattern,
                    time_step=None, overlap=50, n_domains=4, max_level=15,
                    overwrite=False):
    """Stitch all smaller domains into largest domain and save output

    Parameters
    ----------
    year : int
        Year for input files
    month : int
        Month for input files
    time_step : int
        Time step for input files for the specified month. e.g. if year=2017,
        month=3, time_step=0 this will select the file for the first time step
        of 2017-03-01. If None then stitch and save will be done for full
        month.
    input_pattern : str
        Pattern for input files. Assumes pattern contains {month}, {year}, and
        {domain}
    out_pattern : str
        Pattern for output files
    overlap : int
        Number of grid points to use for blending edges
    n_domains : int
        Number of domains to stitch together
    max_level : int
        Max pressure level index
    overwrite : bool
        Whether to overwrite existing files
    """
    logger.info(f'Getting file patterns for year={year}, month={month}')
    input_files, out_files = get_files(year, month, input_pattern,
                                       out_pattern, n_domains=n_domains)
    out_files = (out_files if time_step is None
                 else out_files[time_step - 1: time_step])
    for i, out_file in enumerate(out_files):
        if not os.path.exists(out_file) or overwrite:
            handles = stitch_domains(year, month, i, input_files,
                                     overlap=overlap, n_domains=n_domains,
                                     max_level=max_level)
            basedir = os.path.dirname(out_file)
            os.makedirs(basedir, exist_ok=True)
            handles[0].to_netcdf(out_file)
            logger.info(f'Saved stitched file to {out_file}')



# -*- coding: utf-8 -*-
"""
Sup3r command line interface (CLI).
"""
import click
import logging

from sup3r.version import __version__
from sup3r.pipeline.forward_pass_cli import from_config as fwp_cli
from sup3r.solar.solar_cli import from_config as solar_cli
from sup3r.preprocessing.data_extract_cli import from_config as dh_cli
from sup3r.postprocessing.data_collect_cli import from_config as dc_cli
from sup3r.qa.qa_cli import from_config as qa_cli
from sup3r.qa.visual_qa_cli import from_config as visual_qa_cli
from sup3r.qa.stats_cli import from_config as stats_cli
from sup3r.pipeline.pipeline_cli import from_config as pipe_cli
from sup3r.pipeline.pipeline_cli import valid_config_keys as pipeline_keys
from sup3r.batch.batch_cli import from_config as batch_cli
from sup3r.bias.bias_calc_cli import from_config as bias_calc_cli
from sup3r.utilities.regridder_cli import from_config as regrid_cli


logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--config_file', '-c',
              required=True, type=click.Path(exists=True),
              help='sup3r configuration file json for a single module.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, config_file, verbose):
    """Sup3r command line interface.

    Try using the following commands to pull up the help pages for the
    respective sup3r CLIs::

        $ sup3r --help

        $ sup3r -c config.json pipeline --help

        $ sup3r -c config.json forward-pass --help

        $ sup3r -c config.json data-collect --help

        $ sup3r -c config.json data-extract --help

    Typically, a good place to start is to set up a sup3r job with a pipeline
    config that points to several sup3r modules that you want to run in serial.
    You would call the sup3r pipeline CLI using either of these equivalent
    commands::

        $ sup3r -c config_pipeline.json pipeline

        $ sup3r-pipeline from-config -c config_pipeline.json

    See the help pages of the module CLIs for more details on the config files
    for each CLI.
    """
    ctx.ensure_object(dict)
    ctx.obj['CONFIG_FILE'] = config_file
    ctx.obj['VERBOSE'] = verbose


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def forward_pass(ctx, verbose):
    """Sup3r forward pass to super-resolve data.

    The sup3r forward pass is where all the magic happens. This module takes in
    low resolution data from source files (e.g. WRF or GCM outputs), loads a
    saved Sup3rGan model, does a forward pass of the low res data through the
    generator, and then saves the output high resolution data to disk. This
    module also handles multi-node and multi-core parallelization.

    You can call the forward-pass module via the sup3r-pipeline CLI, or call it
    directly with either of these equivalent commands::

        $ sup3r -c config_fwp.json forward-pass

        $ sup3r-forward-pass from-config -c config_fwp.json

    A sup3r forward pass config.json file can contain any arguments or keyword
    arguments required to initialize the
    :class:`sup3r.pipeline.forward_pass.ForwardPassStrategy` module. The config
    also has several optional arguments: ``log_pattern``, ``log_level``, and
    ``execution_control``. Here's a small example forward pass config::

        {
            "file_paths": "./source_files*.nc",
            "model_kwargs": {
                "model_dir": "./sup3r_model/"
            },
            "out_pattern": "./output/sup3r_out_{file_id}.h5",
            "log_pattern": "./logs/log_{node_index}.log",
            "log_level": "DEBUG",
            "fwp_chunk_shape": [5, 5, 3],
            "spatial_pad": 1,
            "temporal_pad": 1,
            "max_nodes": 1,
            "worker_kwargs": {
                "max_workers": null,
                "output_workers": 1,
                "pass_workers": 8,
                "ti_workers": 1
            },
            "input_handler_kwargs": {
                "worker_kwargs": {
                    "max_workers": 1
                },
            },
            "execution_control": {
                "option": "local"
            },
            "execution_control_eagle": {
                "option": "eagle",
                "walltime": 4,
                "alloc": "sup3r"
            }
        }

    Note that the ``execution_control`` block will run the job locally, while
    the ``execution_control_eagle`` block are kwargs that would be required to
    distribute the job on multiple nodes on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(fwp_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def solar(ctx, verbose):
    """Sup3r solar module to convert GAN output clearsky ratio to irradiance

    Typically we train solar GAN's on clearsky ratio to remove the dependence
    on known variables like solar position and clearsky irradiance. This module
    converts the clearsky ratio output by the GAN in the forward-pass step to
    actual irradiance values (ghi, dni, and dhi). This should be run after the
    forward-pass but before the data-collect step.

    You can call the solar module via the sup3r-pipeline CLI, or call it
    directly with either of these equivalent commands::

        $ sup3r -c config_solar.json solar

        $ sup3r-solar from-config -c config_solar.json

    A sup3r solar config.json file can contain any arguments or keyword
    arguments required to call the
    :meth:`sup3r.solar.solar.Solar.run_temporal_chunk` method. You do not need
    to include the ``i_t_chunk`` input, this is added by the CLI. The config
    also has several optional arguments: ``log_pattern``, ``log_level``, and
    ``execution_control``. Here's a small example solar config::

        {
            "fp_pattern": "./chunks/sup3r*.h5",
            "nsrdb_fp": "/datasets/NSRDB/current/nsrdb_2015.h5",
            "execution_control": {
                "option": "local"
            },
            "execution_control_eagle": {
                "option": "eagle",
                "walltime": 4,
                "alloc": "sup3r"
            }
        }

    Note that the ``execution_control`` block will run the job locally, while
    the ``execution_control_eagle`` block are kwargs that would be required to
    distribute the job on multiple nodes on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(solar_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def bias_calc(ctx, verbose):
    """Sup3r bias correction calculation module to create bias correction
    factors for low res input data.

    This is typically used to understand and correct the biases in global
    climate model (GCM) data from CMIP. Bias correcting GCM data is a very
    common process when dealing with climate data and this helps automate that
    process prior to passing GCM data through sup3r generative models. This is
    typically the first step in a GCM downscaling pipeline.

    You can call the bias calc module via the sup3r-pipeline CLI, or call it
    directly with either of these equivalent commands::

        $ sup3r -c config_bias.json bias-calc

        $ sup3r-bias-calc from-config -c config_bias.json

    A sup3r bias calc config.json file can contain any arguments or keyword
    arguments required to call the

    The config has high level ``bias_calc_class`` and ``jobs`` keys. The
    ``bias_calc_class`` is a class name from the :mod:`sup3r.bias.bias_calc`
    module, and the ``jobs`` argument is a list of kwargs required to
    initialize the ``bias_calc_class`` and run the ``bias_calc_class.run()``
    method (for example, see
    :meth:`sup3r.bias.bias_calc.LinearCorrection.run`). There are also has
    several optional arguments: ``log_pattern``, ``log_level``, and
    ``execution_control``. Here's a small example bias calc config::

        {
            "bias_calc_class": "LinearCorrection",
            "jobs": [
                {
                    "base_fps" : ["/datasets/WIND/HRRR/HRRR_2015.h5"],
                    "bias_fps": ["./ta_day_EC-Earth3-Veg_ssp585.nc"],
                    "base_dset": "U_100m",
                    "bias_feature": "U_100m",
                    "target": [20, -130],
                    "shape": [48, 95]
                 }
            ],
            "execution_control": {
                "option": "local"
            },
            "execution_control_eagle": {
                "option": "eagle",
                "walltime": 4,
                "alloc": "sup3r"
            }
        }

    Note that the ``execution_control`` block will run the job locally, while
    the ``execution_control_eagle`` block are kwargs that would be required to
    distribute the job on multiple nodes on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(bias_calc_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def data_extract(ctx, verbose):
    """Sup3r data extraction and caching prior to training

    The sup3r data-extract module is a utility to pre-extract and pre-process
    data from a source file to disk pickle files for faster restarts while
    debugging. You can call the data-extract module via the sup3r-pipeline CLI,
    or call it directly with either of these equivalent commands::

        $ sup3r -c config_extract.json data-extract

        $ sup3r-data-extract from-config -c config_extract.json

    A sup3r data-extract config.json file can contain any arguments or keyword
    arguments required to initialize the
    :class:`sup3r.preprocessing.data_handling.DataHandler` class. The config
    also has several optional arguments: ``handler_class``, ``log_level``, and
    ``execution_control``. Here's a small example data-extract config::

        {
            "file_paths": "/datasets/WIND/conus/v1.0.0/wtk_conus_2007.h5",
            "features": ["U_100m", "V_100m"],
            "target": [27, -97],
            "shape": [800, 800],
            "execution_control": {"option": "local"},
            "log_level": "DEBUG"
        }

    Note that the ``execution_control`` has the same options as forward-pass
    and you can set ``"option": "eagle"`` to run on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(dh_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def data_collect(ctx, verbose):
    """Sup3r data collection following forward pass.

    The sup3r data-collect module can be used to collect time-chunked files
    that were spit out by multi-node forward pass jobs. You can call the
    data-collect module via the sup3r-pipeline CLI, or call it directly with
    either of these equivalent commands::

        $ sup3r -c config_collect.json data-collect

        $ sup3r-data-collect from-config -c config_collect.json

    A sup3r data-collect config.json file can contain any arguments or keyword
    arguments required to run the
    :meth:`sup3r.postprocessing.collection.Collector.collect` method. The
    config also has several optional arguments: ``log_file``, ``log_level``,
    and ``execution_control``. Here's a small example data-collect config::

        {
            "file_paths": "./outputs/*.h5",
            "out_file": "./outputs/output_file.h5",
            "features": ["windspeed_100m", "winddirection_100m"],
            "log_file": "./logs/collect.log",
            "execution_control": {"option": "local"},
            "log_level": "DEBUG"
        }

    Note that the ``execution_control`` has the same options as forward-pass
    and you can set ``"option": "eagle"`` to run on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(dc_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def qa(ctx, verbose):
    """Sup3r QA module following forward pass and collection.

    The sup3r QA module can be used to verify how well the high-resolution
    sup3r resolved outputs adhere to the low-resolution source data. You can
    call the QA module via the sup3r-pipeline CLI, or call it
    directly with either of these equivalent commands::

        $ sup3r -c config_qa.json qa

        $ sup3r-qa from-config -c config_qa.json

    A sup3r QA config.json file can contain any arguments or keyword
    arguments required to initialize the :class:`sup3r.qa.qa.Sup3rQa` class.
    The config also has several optional arguments: ``log_file``,
    ``log_level``, and ``execution_control``. Here's a small example
    QA config::

        {
            "source_file_paths": "./source_files*.nc",
            "out_file_path": "./outputs/collected_output_file.h5",
            "s_enhance": 25,
            "t_enhance": 24,
            "temporal_coarsening_method": "average",
            "log_file": "./logs/qa.log",
            "execution_control": {"option": "local"},
            "log_level": "DEBUG"
        }

    Note that the ``execution_control`` has the same options as forward-pass
    and you can set ``"option": "eagle"`` to run on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(qa_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def visual_qa(ctx, verbose):
    """Sup3r visual QA module following forward pass and collection.

    The sup3r visual QA module can be used to perform a visual inspection on
    the high-resolution sup3r resolved output. You can call the visual QA
    module via the sup3r-pipeline CLI, or call it directly with either of
    these equivalent commands::

        $ sup3r -c config_qa.json visual-qa

        $ sup3r-visual-qa from-config -c config_qa.json

    A sup3r visual QA config.json file can contain any arguments or keyword
    arguments required to initialize the :class:`sup3r.qa.qa.Sup3rVisualQa`
    class. The config also has several optional arguments: ``log_file``,
    ``log_level``, and ``execution_control``. Here's a small example
    visual QA config::

        {
            "file_paths": "./outputs/collected_output*.h5",
            "out_pattern": "./outputs/plots/{feature}_{index}.png",
            "features": ['windspeed_100m', 'winddirection_100m'],
            "time_step": 100,
            "spatial_slice": [None, None, 100],
            "execution_control": {"option": "local"},
            "log_level": "DEBUG"
        }

    Note that the ``execution_control`` has the same options as forward-pass
    and you can set ``"option": "eagle"`` to run on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(visual_qa_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def stats(ctx, verbose):
    """Sup3r stats module following forward pass and collection.

    The sup3r stats module computes various statistics on wind fields of at
    given hub heights. These statistics include energy spectra, time derivative
    pdfs, velocity gradient pdfs, and vorticity pdfs.
    You can call the stats module via the sup3r-pipeline CLI, or call it
    directly with either of these equivalent commands::

        $ sup3r -c config_stats.json stats

        $ sup3r-stats from-config -c config_stats.json

    A sup3r stats config.json file can contain any arguments or keyword
    arguments required to initialize the
    :class:`sup3r.qa.stats.Sup3rStatsMulti` class. The config also has several
    optional arguments: ``log_file``, ``log_level``, and ``execution_control``.
    Here's a small example stats config::

        {
            "source_file_paths": "./source_files*.nc",
            "out_file_path": "./outputs/collected_output_file.h5",
            "s_enhance": 2,
            "t_enhance": 12,
            "features": ["windspeed_100m", "winddirection_100m"],
            "include_stats": ["time_derivative", "gradient", "spectrum_k"]
            "get_interp": True,
            "log_file": "./logs/stats.log",
            "execution_control": {"option": "local"},
            "log_level": "DEBUG"
        }

    Note that the ``execution_control`` has the same options as forward-pass
    and you can set ``"option": "eagle"`` to run on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(stats_cli, config_file=config_file, verbose=verbose)


@main.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def regrid(ctx, verbose):
    """Sup3r regrid module for regridding forward pass output to a different
    meta file.

    You can call the regrid module via the sup3r-pipeline CLI, or call it
    directly with either of these equivalent commands::

        $ sup3r -c config_regrid.json regrid

        $ sup3r-regrid from-config -c config_regrid.json

    A sup3r regrid config.json file can contain any arguments or keyword
    arguments required to initialize the
    :class:`sup3r.utilities.regridder.RegridOutput` class. The config also has
    several optional arguments: ``log_file``, ``log_level``, and
    ``execution_control``.
    Here's a small example regrid config::

        {
            "source_files": "./source_files*.h5",
            "out_pattern": "./chunks_{file_id}.h5",
            "heights": [100, 200],
            "target_meta": "./target_meta.csv",
            "n_chunks": 100,
            "worker_kwargs": {"regrid_workers": 10, "query_workers": 10},
            "cache_pattern": "./{array_name}.pkl",
            "log_file": "./logs/regrid.log",
            "execution_control": {"option": "local"},
            "log_level": "DEBUG"
        }

    Note that the ``execution_control`` has the same options as forward-pass
    and you can set ``"option": "eagle"`` to run on the NREL HPC.
    """
    config_file = ctx.obj['CONFIG_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])
    ctx.invoke(regrid_cli, config_file=config_file, verbose=verbose)


@main.group(invoke_without_command=True)
@click.option('--cancel', is_flag=True,
              help='Flag to cancel all jobs associated with a given pipeline.')
@click.option('--monitor', is_flag=True,
              help='Flag to monitor pipeline jobs continuously. '
              'Default is not to monitor (kick off jobs and exit).')
@click.option('--background', is_flag=True,
              help='Flag to monitor pipeline jobs continuously '
              'in the background using the nohup command. Note that the '
              'stdout/stderr will not be captured, but you can set a '
              'pipeline "log_file" to capture logs.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def pipeline(ctx, cancel, monitor, background, verbose):
    """Execute multiple steps in a Sup3r pipeline.

    Typically, a good place to start is to set up a sup3r job with a pipeline
    config that points to several sup3r modules that you want to run in serial.
    You would call the sup3r pipeline CLI using either of these equivalent
    commands::

        $ sup3r -c config_pipeline.json pipeline

        $ sup3r-pipeline from-config -c config_pipeline.json

    A typical sup3r pipeline config.json file might look like this::

        {
            "logging": {"log_level": "DEBUG"},
            "pipeline": [
                {"forward-pass": "./config_fwp.json"},
                {"data-collect": "./config_collect.json"}
            ]
        }

    See the other CLI help pages for what the respective module configs
    require.
    """
    if ctx.invoked_subcommand is None:
        config_file = ctx.obj['CONFIG_FILE']
        verbose = any([verbose, ctx.obj['VERBOSE']])
        ctx.invoke(pipe_cli, config_file=config_file, cancel=cancel,
                   monitor=monitor, background=background, verbose=verbose)


@pipeline.command()
@click.pass_context
def valid_pipeline_keys(ctx):
    """Print the valid pipeline config keys"""
    ctx.invoke(pipeline_keys)


@main.group(invoke_without_command=True)
@click.option('--dry-run', is_flag=True,
              help='Flag to do a dry run (make batch dirs without running).')
@click.option('--cancel', is_flag=True,
              help='Flag to cancel all jobs associated with a given batch.')
@click.option('--delete', is_flag=True,
              help='Flag to delete all batch job sub directories associated '
              'with the batch_jobs.csv in the current batch config directory.')
@click.option('--monitor-background', is_flag=True,
              help='Flag to monitor all batch pipelines continuously '
              'in the background using the nohup command. Note that the '
              'stdout/stderr will not be captured, but you can set a '
              'pipeline "log_file" to capture logs.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def batch(ctx, dry_run, cancel, delete, monitor_background, verbose):
    """Create and run multiple sup3r project directories based on batch
    permutation logic.

    The sup3r batch module (built on the reV batch functionality) is a way to
    create and run many sup3r pipeline projects based on permutations of
    key-value pairs in the run config files. A user configures the batch file
    by creating one or more "sets" that contain one or more arguments (keys
    found in config files) that are to be parameterized. For example, in the
    config below, four sup3r pipelines will be created where arg1 and arg2 are
    set to [0, "a"], [0, "b"], [1, "a"], [1, "b"] in config_fwp.json::

        {
            "pipeline_config": "./config_pipeline.json",
            "sets": [
              {
                "args": {
                  "arg1": [0, 1],
                  "arg2": ["a", "b"],
                },
                "files": ["./config_fwp.json"],
                "set_tag": "set1"
              }
        }

    Note that you can use multiple "sets" to isolate parameter permutations.
    """
    if ctx.invoked_subcommand is None:
        config_file = ctx.obj['CONFIG_FILE']
        verbose = any([verbose, ctx.obj['VERBOSE']])
        ctx.invoke(batch_cli, config_file=config_file,
                   dry_run=dry_run, cancel=cancel, delete=delete,
                   monitor_background=monitor_background,
                   verbose=verbose)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running sup3r CLI')
        raise



# -*- coding: utf-8 -*-
"""
sup3r QA module CLI entry points.
"""
import click
import logging

from sup3r.utilities import ModuleName
from sup3r.version import __version__
from sup3r.qa.qa import Sup3rQa
from sup3r.utilities.cli import BaseCLI


logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, verbose):
    """Sup3r QA module Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='sup3r QA configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run the sup3r QA module from a config file."""
    BaseCLI.from_config(ModuleName.QA, Sup3rQa, ctx, config_file, verbose)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running sup3r QA CLI')
        raise
